# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import warnings

import torch
import triton
import triton.language as tl

from fla.ops.utils import prepare_chunk_indices
from fla.ops.utils.cumsum import chunk_global_cumsum, chunk_local_cumsum
from fla.ops.utils.op import exp
from fla.utils import (
    IS_INTEL_ALCHEMIST,
    IS_NVIDIA_HOPPER,
    autocast_custom_bwd,
    autocast_custom_fwd,
    autotune_cache_kwargs,
    check_shared_mem,
    input_guard,
)

# https://github.com/intel/intel-xpu-backend-for-triton/issues/3449
triton_config = {'grf_mode': 'large'} if IS_INTEL_ALCHEMIST else {}
NUM_WARPS = [2, 4, 8] if IS_NVIDIA_HOPPER else [2, 4, 8, 16]


@triton.heuristics({
    'NV': lambda args: triton.cdiv(args['V'], args['BV']),
    'OUTPUT_ATTENTIONS': lambda args: args['attn'] is not None,
    'USE_G': lambda args: args['g'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [2, 4, 8, 16]
        for num_stages in [2, 3, 4]
    ],
    key=["BT", "BS", "BK", "BV", "USE_G"],
    **autotune_cache_kwargs,
)
@triton.jit
def parallel_simple_gla_fwd_kernel(
    q,
    k,
    v,
    g,
    o,
    attn,
    scale,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    NV: tl.constexpr,
    OUTPUT_ATTENTIONS: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    USE_G: tl.constexpr,
):
    i_kv, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_k, i_v = i_kv // NV, i_kv % NV
    i_b, i_h = i_bh // H, i_bh % H

    all = B * T
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    q += (bos * H + i_h) * K
    k += (bos * H + i_h) * K
    v += (bos * H + i_h) * V
    o += ((i_k * all + bos) * H + i_h) * V
    if USE_G:
        g += bos * H + i_h
    if OUTPUT_ATTENTIONS:
        attn += i_k * B * H * T * T + (bos * H + i_h * T) * T

    p_q = tl.make_block_ptr(q, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))

    # the Q block is kept in the shared memory throughout the whole kernel
    # [BT, BK]
    b_q = tl.load(p_q, boundary_check=(0, 1))
    b_q = (b_q * scale).to(b_q.dtype)
    b_o = tl.zeros([BT, BV], dtype=tl.float32)

    # [BT]
    o_q = i_t * BT + tl.arange(0, BT)
    m_q = o_q < T
    # Q block and K block have overlap.
    # masks required
    if USE_G:
        # [BT,]
        b_gq = tl.load(g + o_q * H, mask=m_q, other=float('-inf')).to(tl.float32)
        # rescale interchunk output
    else:
        b_gq = None

    for i_s in range(i_t * BT, min((i_t + 1) * BT, T), BS):
        p_k = tl.make_block_ptr(k, (K, T), (1, H*K), (i_k * BK, i_s), (BK, BS), (0, 1))
        p_v = tl.make_block_ptr(v, (T, V), (H*V, 1), (i_s, i_v * BV), (BS, BV), (1, 0))

        o_k = i_s + tl.arange(0, BS)
        m_k = o_k < T
        # [BK, BS]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [BS, BV]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BT, BS]
        m_s = (o_q[:, None] >= o_k[None, :]) & (m_q[:, None] & m_k[None, :])
        b_s = tl.dot(b_q, b_k)
        if USE_G:
            b_gk = tl.load(g + o_k * H, mask=m_k, other=0)
            b_s *= exp(b_gq[:, None] - b_gk[None, :])
        b_s = tl.where(m_s, b_s, 0)
        # [BT, BV]
        if i_s >= 0:
            b_o += tl.dot(b_s.to(b_q.dtype), b_v)
        if OUTPUT_ATTENTIONS:
            p_a = tl.make_block_ptr(attn, (T, T), (T, 1), (i_t * BT, i_s), (BT, BS), (1, 0))
            tl.store(p_a, b_s.to(p_a.dtype.element_ty), boundary_check=(0, 1))
    for i_s in range(i_t * BT - BS, -BS, -BS):
        p_k = tl.make_block_ptr(k, (K, T), (1, H*K), (i_k * BK, i_s), (BK, BS), (0, 1))
        p_v = tl.make_block_ptr(v, (T, V), (H*V, 1), (i_s, i_v * BV), (BS, BV), (1, 0))

        o_k = i_s + tl.arange(0, BS)
        m_k = o_k < T
        # [BK, BS]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [BS, BV]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BT, BS]
        m_s = m_q[:, None] & m_k[None, :]
        b_s = tl.dot(b_q, b_k)
        if USE_G:
            b_g = tl.load(g + o_k * H, mask=m_k, other=0)
            b_gn = tl.load(g + (min(i_s + BS, T) - 1) * H)
            b_gp = tl.load(g + (i_s-1) * H) if i_s % BT > 0 else 0.
            # No concrete meaning. Just to avoid some layout bugs.
            b_s *= exp(b_gq[:, None] + (b_gn - b_g)[None, :])
            b_gq += b_gn - b_gp
        b_s = tl.where(m_s, b_s, 0)
        if OUTPUT_ATTENTIONS:
            p_a = tl.make_block_ptr(attn, (T, T), (T, 1), (i_t * BT, i_s), (BT, BS), (1, 0))
            tl.store(p_a, b_s.to(p_a.dtype.element_ty), boundary_check=(0, 1))
        if i_s >= 0:
            b_o += tl.dot(b_s.to(b_v.dtype), b_v)
    p_o = tl.make_block_ptr(o, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))


@triton.jit(do_not_specialize=['T'])
def parallel_simple_gla_bwd_kernel_dq(
    i_t,
    i_k,
    i_v,
    q,
    k,
    v,
    g,
    do,
    dq,
    dg,
    scale,
    T,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    USE_G: tl.constexpr,
):
    p_do = tl.make_block_ptr(do, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    # [BT, BV]
    b_do = tl.load(p_do, boundary_check=(0, 1))
    # [BT, BK]
    b_dq = tl.zeros([BT, BK], dtype=tl.float32)

    # [BT]
    o_q = i_t * BT + tl.arange(0, BT)
    m_q = o_q < T
    for i_s in range(0, i_t * BT, BS):
        p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (i_s, i_k * BK), (BS, BK), (1, 0))
        p_v = tl.make_block_ptr(v, (V, T), (1, H*V), (i_v * BV, i_s), (BV, BS), (0, 1))

        o_k = i_s + tl.arange(0, BS)
        m_k = o_k < T
        # [BS, BK]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [BV, BS]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BT, BV] @ [BV, BS] = [BT, BS]
        b_ds = tl.dot(b_do, b_v)
        if USE_G:
            b_g = tl.load(g + o_k * H, mask=m_k, other=0)
            b_gn = tl.load(g + (min(i_s + BS, T) - 1) * H)
            b_gp = tl.load(g + (i_s - 1) * H) if i_s % BT > 0 else 0.
            b_ds *= tl.where(m_k, exp(b_gn - b_g), 0)[None, :]
            if i_s > 0:
                b_dq *= exp(b_gn - b_gp)
        # [BT, BS] @ [BS, BK] = [BT, BK]
        b_dq += tl.dot(b_ds.to(b_v.dtype), b_k)

    if USE_G:
        # [BT,]
        b_gq = tl.load(g + o_q * H, mask=m_q, other=float('-inf'))
        # [BT, BK]
        b_dq *= exp(b_gq)[:, None]

    # Q block and K block have overlap. masks required
    for i_s in range(i_t * BT, min((i_t + 1) * BT, T), BS):
        p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (i_s, i_k * BK), (BS, BK), (1, 0))
        p_v = tl.make_block_ptr(v, (V, T), (1, H*V), (i_v * BV, i_s), (BV, BS), (0, 1))

        o_k = i_s + tl.arange(0, BS)
        m_k = o_k < T
        # [BS, BK]
        b_k = tl.load(p_k, boundary_check=(0, 1))
        # [BV, BS]
        b_v = tl.load(p_v, boundary_check=(0, 1))
        # [BT, BV] @ [BV, BS] = [BT, BS]
        b_ds = tl.dot(b_do, b_v)
        if USE_G:
            b_gk = tl.load(g + o_k * H, mask=m_k, other=0)
            b_ds *= exp(b_gq[:, None] - b_gk[None, :])
        m_s = (o_q[:, None] >= o_k[None, :]) & (m_q[:, None] & m_k[None, :])
        b_ds = tl.where(m_s, b_ds, 0)
        # [BT, BK]
        b_dq += tl.dot(b_ds.to(b_k.dtype), b_k)

    b_dq *= scale
    p_dq = tl.make_block_ptr(dq, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
    tl.store(p_dq, b_dq.to(p_dq.dtype.element_ty), boundary_check=(0, 1))
    if USE_G:
        p_q = tl.make_block_ptr(q, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
        b_q = tl.load(p_q, boundary_check=(0, 1))
        b_dg = tl.sum(b_dq * b_q, 1)
        p_dg = tl.make_block_ptr(dg, (T,), (H,), (i_t * BT,), (BT,), (0,))
        tl.store(p_dg, b_dg.to(p_dg.dtype.element_ty), boundary_check=(0,))


@triton.jit(do_not_specialize=['T'])
def parallel_simple_gla_bwd_kernel_dkv(
    i_t,
    i_k,
    i_v,
    q,
    k,
    v,
    g,
    do,
    dk,
    dv,
    dg,
    scale,
    T,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    USE_G: tl.constexpr,
):
    o_k = i_t * BT + tl.arange(0, BT)
    m_k = o_k < T
    # [BT, BK]
    p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
    b_k = tl.load(p_k, boundary_check=(0, 1))
    b_dk = tl.zeros([BT, BK], dtype=tl.float32)
    # [BT, BV]
    p_v = tl.make_block_ptr(v, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    b_v = tl.load(p_v, boundary_check=(0, 1))
    b_dv = tl.zeros([BT, BV], dtype=tl.float32)
    if USE_G:
        b_gk = tl.load(g + o_k * H, mask=m_k, other=0)
    NTS = tl.cdiv(T, BS)
    # [BT, BK]
    for i_s in range(NTS * BS - BS, (i_t + 1) * BT - BS, -BS):
        p_q = tl.make_block_ptr(q, (T, K), (H*K, 1), (i_s, i_k * BK), (BS, BK), (1, 0))
        p_do = tl.make_block_ptr(do, (T, V), (H*V, 1), (i_s, i_v * BV), (BS, BV), (1, 0))

        o_q = i_s + tl.arange(0, BS)
        m_q = o_q < T
        # [BS, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        # [BS, BV]
        b_do = tl.load(p_do, boundary_check=(0, 1))
        # [BT, BS]
        b_ds = tl.dot(b_v, tl.trans(b_do))
        b_s = tl.dot(b_k, tl.trans(b_q))
        if USE_G:
            b_gq = tl.load(g + o_q * H, mask=m_q, other=float('-inf'))
            b_gp = tl.load(g + (min(i_s + BS, T) - 1) * H)
            b_gn = tl.load(g + (i_s - 1) * H) if i_s % BT > 0 else 0.
            if i_s >= 0:
                b_gpn = exp(b_gp - b_gn)
                b_dk *= b_gpn
                b_dv *= b_gpn
                b_gqn = exp(b_gq - b_gn)
                b_ds *= b_gqn[None, :]
                b_s *= b_gqn[None, :]
        # [BT, BK]
        b_dk += tl.dot(b_ds.to(b_q.dtype), b_q)
        # [BT, BV]
        b_dv += tl.dot(b_s.to(b_do.dtype), b_do)

    if USE_G:
        b_gn = tl.load(g + (min(i_t * BT + BT, T) - 1) * H)
        if i_t >= 0:
            b_gpn = exp(b_gn - b_gk)[:, None]
            b_dk *= b_gpn
            b_dv *= b_gpn

    for i_s in range(i_t * BT, min((i_t + 1) * BT, T), BS):
        p_q = tl.make_block_ptr(q, (T, K), (H*K, 1), (i_s, i_k * BK), (BS, BK), (1, 0))
        p_do = tl.make_block_ptr(do, (T, V), (H*V, 1), (i_s, i_v * BV), (BS, BV), (1, 0))

        o_q = i_s + tl.arange(0, BS)
        m_q = o_q < T
        # [BS, BK]
        b_q = tl.load(p_q, boundary_check=(0, 1))
        # [BS, BV]
        b_do = tl.load(p_do, boundary_check=(0, 1))
        # [BS]
        b_s = tl.dot(b_k, tl.trans(b_q))
        b_ds = tl.dot(b_v, tl.trans(b_do))
        if USE_G:
            b_gq = tl.load(g + o_q * H, mask=m_q, other=float('-inf'))
            if i_s >= 0:
                b_gkq = exp(-b_gk[:, None] + b_gq[None, :])
                b_ds *= b_gkq
                b_s *= b_gkq
        m_s = o_k[:, None] <= o_q[None, :]
        b_s = tl.where(m_s, b_s, 0)
        b_ds = tl.where(m_s, b_ds, 0)
        # [BT, BK]
        b_dk += tl.dot(b_ds.to(b_q.dtype), b_q)
        b_dv += tl.dot(b_s.to(b_do.dtype), b_do)
    b_dk *= scale
    b_dv *= scale
    p_dk = tl.make_block_ptr(dk, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
    p_dv = tl.make_block_ptr(dv, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
    tl.store(p_dk, b_dk.to(p_dk.dtype.element_ty), boundary_check=(0, 1))
    tl.store(p_dv, b_dv.to(p_dv.dtype.element_ty), boundary_check=(0, 1))
    if USE_G:
        b_dg = tl.load(dg + o_k * H, mask=m_k, other=0)
        b_dg -= tl.sum(b_dk * b_k, 1)
        tl.store(dg + o_k * H, b_dg.to(dg.dtype.element_ty), mask=m_k)


@triton.heuristics({
    'NV': lambda args: triton.cdiv(args['V'], args['BV']),
    'USE_G': lambda args: args['g'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config(triton_config, num_warps=num_warps)
        for num_warps in NUM_WARPS
    ],
    key=['BT', 'BS', 'BK', 'BV', 'USE_G'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def parallel_simple_gla_bwd_kernel(
    q,
    k,
    v,
    g,
    do,
    dq,
    dk,
    dv,
    dg,
    scale,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    NV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    USE_G: tl.constexpr,
):
    i_kv, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_k, i_v = i_kv // NV, i_kv % NV
    i_b, i_h = i_bh // H, i_bh % H
    dq += i_v * B * H * T * K
    dk += i_v * B * H * T * K
    dv += i_k * B * H * T * V
    if USE_G:
        dg += i_kv * B * H * T

    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    q += (bos * H + i_h) * K
    k += (bos * H + i_h) * K
    v += (bos * H + i_h) * V
    do += (bos * H + i_h) * V
    dq += (bos * H + i_h) * K
    dk += (bos * H + i_h) * K
    dv += (bos * H + i_h) * V
    if USE_G:
        g += bos * H + i_h
        dg += bos * H + i_h

    parallel_simple_gla_bwd_kernel_dq(
        i_t=i_t,
        i_k=i_k,
        i_v=i_v,
        q=q,
        k=k,
        v=v,
        g=g,
        do=do,
        dq=dq,
        dg=dg,
        scale=scale,
        T=T,
        H=H,
        K=K,
        V=V,
        BT=BT,
        BS=BS,
        BK=BK,
        BV=BV,
        USE_G=USE_G,
    )
    tl.debug_barrier()
    parallel_simple_gla_bwd_kernel_dkv(
        i_t=i_t,
        i_k=i_k,
        i_v=i_v,
        q=q,
        k=k,
        v=v,
        g=g,
        do=do,
        dk=dk,
        dv=dv,
        dg=dg,
        scale=scale,
        T=T,
        H=H,
        K=K,
        V=V,
        BT=BT,
        BS=BS,
        BK=BK,
        BV=BV,
        USE_G=USE_G,
    )


def parallel_simple_gla_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    scale: float,
    output_attentions: bool = False,
    chunk_size: int = 128,
    cu_seqlens: torch.LongTensor | None = None,
):
    B, T, H, K, V = *k.shape, v.shape[-1]
    BT, BS = chunk_size, 32
    if check_shared_mem('hopper', k.device.index):
        BK = min(256, triton.next_power_of_2(K))
        BV = min(256, triton.next_power_of_2(V))
    elif check_shared_mem('ampere', k.device.index):
        BK = min(128, triton.next_power_of_2(K))
        BV = min(128, triton.next_power_of_2(V))
    else:
        BK = min(64, triton.next_power_of_2(K))
        BV = min(64, triton.next_power_of_2(V))

    NK = triton.cdiv(K, BK)
    NV = triton.cdiv(V, BV)
    assert BT % BS == 0

    chunk_indices = prepare_chunk_indices(cu_seqlens, chunk_size) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)

    # local cumulative decay in log space
    if g is not None:
        g = chunk_local_cumsum(g, chunk_size, cu_seqlens=cu_seqlens)
    grid = (NK * NV, NT, B * H)
    o = torch.empty(NK, *v.shape, dtype=v.dtype if NK == 1 else torch.float, device=q.device)
    attn = q.new_zeros(NK, B, H, T, T) if output_attentions else None

    parallel_simple_gla_fwd_kernel[grid](
        q=q,
        k=k,
        v=v,
        g=g,
        o=o,
        attn=attn,
        scale=scale,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        B=B,
        H=H,
        T=T,
        K=K,
        V=V,
        BT=BT,
        BS=BS,
        BK=BK,
        BV=BV,
    )
    o = o.sum(0)

    if output_attentions:
        attn = attn.sum(0)
    return o, g, attn


def parallel_simple_gla_bwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    do: torch.Tensor,
    scale: float,
    chunk_size: int = 128,
    cu_seqlens: torch.LongTensor | None = None,
):
    B, T, H, K, V = *k.shape, v.shape[-1]
    BT, BS = chunk_size, 32
    if check_shared_mem('hopper', k.device.index):
        BK = min(256, triton.next_power_of_2(K))
        BV = min(256, triton.next_power_of_2(V))
    elif check_shared_mem('ampere', k.device.index):
        BK = min(128, triton.next_power_of_2(K))
        BV = min(128, triton.next_power_of_2(V))
    elif check_shared_mem('ada', k.device.index):
        BK = min(64, triton.next_power_of_2(K))
        BV = min(64, triton.next_power_of_2(V))
    else:
        BK = min(32, triton.next_power_of_2(K))
        BV = min(32, triton.next_power_of_2(V))

    NK = triton.cdiv(K, BK)
    NV = triton.cdiv(V, BV)
    assert BT % BS == 0

    dq = torch.empty(NV, * q.shape, dtype=q.dtype if NV == 1 else torch.float, device=q.device)
    dk = torch.empty(NV, * k.shape, dtype=k.dtype if NV == 1 else torch.float, device=q.device)
    dv = torch.empty(NK, * v.shape, dtype=v.dtype if NK == 1 else torch.float, device=q.device)
    dg = torch.empty(NK*NV, *g.shape, dtype=torch.float, device=q.device) if g is not None else None

    chunk_indices = prepare_chunk_indices(cu_seqlens, chunk_size) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)

    grid = (NK * NV, NT, B * H)
    parallel_simple_gla_bwd_kernel[grid](
        q=q,
        k=k,
        v=v,
        g=g,
        do=do,
        dq=dq,
        dk=dk,
        dv=dv,
        dg=dg,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        scale=scale,
        T=T,
        B=B,
        H=H,
        K=K,
        V=V,
        BT=BT,
        BS=BS,
        BK=BK,
        BV=BV,
    )
    dq = dq.sum(0)
    dk = dk.sum(0)
    dv = dv.sum(0)
    dg = chunk_global_cumsum(dg.sum(0), reverse=True, cu_seqlens=cu_seqlens) if g is not None else None
    return dq, dk, dv, dg


class ParallelSimpleGLAFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    @autocast_custom_fwd
    def forward(ctx, q, k, v, g, scale, output_attentions, cu_seqlens):
        chunk_size = 128
        ctx.dtype = q.dtype

        o, g, attn = parallel_simple_gla_fwd(
            q=q,
            k=k,
            v=v,
            g=g,
            scale=scale,
            output_attentions=output_attentions,
            chunk_size=chunk_size,
            cu_seqlens=cu_seqlens,
        )
        ctx.save_for_backward(q, k, v, g, cu_seqlens)
        ctx.scale = scale
        ctx.chunk_size = chunk_size
        return o.to(q.dtype), attn

    @staticmethod
    @input_guard
    @autocast_custom_bwd
    def backward(ctx, do, da=None):
        q, k, v, g, cu_seqlens = ctx.saved_tensors
        dq, dk, dv, dg = parallel_simple_gla_bwd(
            q=q,
            k=k,
            v=v,
            g=g,
            do=do,
            scale=ctx.scale,
            chunk_size=ctx.chunk_size,
            cu_seqlens=cu_seqlens,
        )
        return dq.to(q), dk.to(k), dv.to(v), dg.to(ctx.dtype) if dg is not None else None, None, None, None


def parallel_simple_gla(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor | None = None,
    scale: float | None = None,
    output_attentions: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
    head_first: bool = False,
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, H, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        g (torch.Tensor):
            Forget gates of shape `[B, T, H]`.
            Compared to GLA, the gating is head-wise instead of elementwise.
        scale (Optional[float]):
            Scale factor for attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        output_attentions (bool):
            Whether to output the materialized attention scores of shape [B, H, T, T]. Default: `False`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.
        head_first (Optional[bool]):
            Whether the inputs are in the head-first format. Default: `False`.
            This argument has been deprecated.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, H, V]`.
        attn (torch.Tensor):
            Attention scores of shape `[B, H, T, T]` if `output_attentions=True` else `None`
    """
    if head_first:
        raise DeprecationWarning(
            "head_first is deprecated and will be removed in a future version. "
            "Please use head_first=False for now instead.",
        )
    if not head_first and q.shape[1] < q.shape[2]:
        warnings.warn(
            f"Input tensor shape suggests potential format mismatch: seq_len ({q.shape[1]}) < num_heads ({q.shape[2]}). "
            "This may indicate the inputs were passed in head-first format [B, H, T, ...] "
            "when head_first=False was specified. "
            "Please verify your input tensor format matches the expected shape [B, T, H, ...].",
        )
    if cu_seqlens is not None:
        if q.shape[0] != 1:
            raise ValueError(
                f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
                f"Please flatten variable-length inputs before processing.",
            )
    if output_attentions:
        assert cu_seqlens is None, "output_attentions=True is not supported with variable-length sequences"

    if scale is None:
        scale = k.shape[-1] ** -0.5
    o, attn = ParallelSimpleGLAFunction.apply(
        q,
        k,
        v,
        g,
        scale,
        output_attentions,
        cu_seqlens,
    )
    return o, attn



"""
ServingDaemon — Long-lived process that holds the InferenceEngine.

Listens on Unix domain socket for JSON-RPC requests.
Single-user, single-connection for V1.

Supports background daemonization: fork to background, parent waits
for ready signal, prints status, exits. Single terminal workflow:
    neurobrix serve → neurobrix chat → neurobrix stop

ZERO HARDCODE: Model name and hardware profile from CLI args.
"""

import os
import sys
import signal
import socket
import traceback
from typing import Any, Dict, Optional

from neurobrix.serving.engine import InferenceEngine
from neurobrix.serving.protocol import (
    SOCKET_PATH, PID_PATH, LOG_PATH, DAEMON_DIR,
    send_message, recv_message,
    make_response,
)


class ServingDaemon:
    """
    Long-running process that holds InferenceEngine with weights in VRAM.
    Accepts connections via Unix socket, dispatches to engine methods.
    """

    def __init__(
        self,
        model_name: str,
        hardware_id: str,
        mode: str = "compiled",
        idle_timeout: float = 1800.0,
    ):
        self._engine = InferenceEngine(model_name, hardware_id, mode)
        self._idle_timeout = idle_timeout
        self._running = False
        self._server_socket: Optional[socket.socket] = None
        self._ready_fd: Optional[int] = None  # Write end of pipe for ready signal

    def start(self, foreground: bool = False) -> None:
        """
        Load model and start serving.

        If foreground=True, runs in current process (blocks terminal).
        If foreground=False (default), forks to background and signals parent when ready.
        """
        if not foreground:
            self._daemonize()
            # After _daemonize(), we are in the child process
            # stdout/stderr redirected to LOG_PATH

        self._start_serving()

    def _daemonize(self) -> None:
        """
        Fork to background. Parent waits for ready signal, then exits.
        Child continues execution with stdout/stderr → LOG_PATH.

        Uses a pipe for clean ready signaling (no polling).
        """
        # Create pipe: parent reads, child writes "ready" when model loaded
        read_fd, write_fd = os.pipe()

        pid = os.fork()

        if pid > 0:
            # ── PARENT PROCESS ──
            os.close(write_fd)

            # Wait for child to signal ready (or die)
            print(f"[Serve] Loading model in background (PID {pid})...")
            print(f"[Serve] Log: {LOG_PATH}")

            read_pipe = os.fdopen(read_fd, 'r')
            try:
                # Block until child writes or pipe closes (child died)
                msg = read_pipe.read()
            finally:
                read_pipe.close()

            if msg.startswith("ready:"):
                # Child loaded successfully
                info = msg[6:].strip()
                print(f"[Serve] {info}")
                print(f"[Serve] Use 'neurobrix chat' to connect")
                print(f"[Serve] Use 'neurobrix stop' to shutdown")
                sys.exit(0)
            elif msg.startswith("error:"):
                print(f"[Serve] Failed: {msg[6:].strip()}")
                sys.exit(1)
            else:
                # Pipe closed without message → child crashed
                print(f"[Serve] Daemon process died unexpectedly. Check {LOG_PATH}")
                sys.exit(1)

        else:
            # ── CHILD PROCESS (daemon) ──
            os.close(read_fd)
            self._ready_fd = write_fd

            # Detach from terminal (new session)
            os.setsid()

            # Redirect stdout/stderr to log file
            DAEMON_DIR.mkdir(parents=True, exist_ok=True)
            log_fd = os.open(str(LOG_PATH), os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o644)
            os.dup2(log_fd, sys.stdout.fileno())
            os.dup2(log_fd, sys.stderr.fileno())
            os.close(log_fd)

            # Redirect stdin from /dev/null
            devnull = os.open(os.devnull, os.O_RDONLY)
            os.dup2(devnull, sys.stdin.fileno())
            os.close(devnull)

    def _signal_ready(self, message: str) -> None:
        """Signal parent process that daemon is ready (or failed)."""
        if self._ready_fd is not None:
            try:
                os.write(self._ready_fd, message.encode())
                os.close(self._ready_fd)
            except OSError:
                pass
            self._ready_fd = None

    def _start_serving(self) -> None:
        """Load model, bind socket, enter serve loop."""
        # Signal handling
        signal.signal(signal.SIGTERM, self._handle_signal)
        signal.signal(signal.SIGINT, self._handle_signal)

        # Ensure daemon directory exists
        DAEMON_DIR.mkdir(parents=True, exist_ok=True)

        # Clean stale socket
        if SOCKET_PATH.exists():
            SOCKET_PATH.unlink()

        # Write PID
        PID_PATH.write_text(str(os.getpid()))

        try:
            # Load model (weights to VRAM)
            self._engine.load()

            # Bind socket
            self._server_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._server_socket.bind(str(SOCKET_PATH))
            self._server_socket.listen(1)
            self._server_socket.settimeout(self._idle_timeout)

            strategy = "warm" if self._engine._warm_serving else "cold"
            family = self._engine.family or "unknown"
            ready_msg = (
                f"ready:Model loaded — {self._engine.model_name} "
                f"({family}, {strategy} serving)"
            )
            self._signal_ready(ready_msg)

            print(f"\n[Daemon] Listening on {SOCKET_PATH}")
            print(f"[Daemon] Idle timeout: {self._idle_timeout}s")

            self._running = True
            self._serve_loop()

        except Exception as e:
            self._signal_ready(f"error:{e}")
            raise

        finally:
            self._cleanup()

    def _serve_loop(self) -> None:
        """Accept connections and handle requests."""
        while self._running:
            try:
                conn, _ = self._server_socket.accept()
            except socket.timeout:
                print(f"[Daemon] Idle timeout ({self._idle_timeout}s) — shutting down")
                break
            except OSError:
                if self._running:
                    print("[Daemon] Socket error")
                break

            try:
                self._handle_connection(conn)
            except Exception as e:
                print(f"[Daemon] Connection error: {e}")
                traceback.print_exc()
            finally:
                conn.close()

    def _handle_connection(self, conn: socket.socket) -> None:
        """Handle a single client connection (may have multiple requests)."""
        conn.settimeout(None)  # No timeout on established connections

        while self._running:
            request = recv_message(conn)
            if request is None:
                break  # Client disconnected

            response = self._dispatch(request)
            send_message(conn, response)

            # Check if shutdown was requested
            if request.get("method") == "shutdown":
                self._running = False
                break

    def _dispatch(self, request: Dict[str, Any]) -> Dict[str, Any]:
        """Dispatch JSON-RPC request to engine method."""
        method = request.get("method", "")
        params = request.get("params", {})

        try:
            if method == "generate":
                output_path = params.pop("output_path", None)
                result = self._engine.generate(**params)
                # For non-LLM: save output file, return path instead of raw tensors
                if output_path and self._engine.family != "llm" and "outputs" in result:
                    saved = self._engine.save_output(result["outputs"], output_path)
                    result.pop("outputs", None)
                    result["output_path"] = saved
                elif "outputs" in result:
                    # Can't serialize raw tensors — drop them
                    result.pop("outputs", None)
                return make_response(result)

            elif method == "chat":
                message = params.get("message", "")
                kwargs = {k: v for k, v in params.items() if k != "message"}
                # Track summarization count before call
                prev_count = 0
                if self._engine._session is not None:
                    prev_count = self._engine._session._summarization_count
                text = self._engine.chat(message, **kwargs)
                context = {}
                summarized = False
                if self._engine._session is not None:
                    context = self._engine._session.get_context_info()
                    summarized = self._engine._session._summarization_count > prev_count
                return make_response({"text": text, "context": context, "summarized": summarized})

            elif method == "new_chat":
                self._engine.new_conversation()
                return make_response({"status": "ok"})

            elif method == "status":
                return make_response(self._engine.get_status())

            elif method == "shutdown":
                self._engine.unload()
                return make_response({"status": "ok"})

            else:
                return make_response(error=f"Unknown method: {method}")

        except Exception as e:
            traceback.print_exc()
            return make_response(error=str(e))

    def _handle_signal(self, signum, frame):
        """Graceful shutdown on SIGTERM/SIGINT."""
        sig_name = signal.Signals(signum).name
        print(f"\n[Daemon] Received {sig_name} — shutting down")
        self._running = False
        # Close server socket to unblock accept()
        if self._server_socket is not None:
            try:
                self._server_socket.close()
            except Exception:
                pass

    def _cleanup(self) -> None:
        """Clean up socket and PID files."""
        if self._server_socket is not None:
            try:
                self._server_socket.close()
            except Exception:
                pass

        if SOCKET_PATH.exists():
            try:
                SOCKET_PATH.unlink()
            except Exception:
                pass

        if PID_PATH.exists():
            try:
                PID_PATH.unlink()
            except Exception:
                pass

        print("[Daemon] Cleaned up")



import logging
import os

import triton
import triton.language as tl
from _kunlunxin.utils.codegen_config_utils import CodeGenConfig

from flag_gems.runtime import device

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
device = device.name

config_ = CodeGenConfig(
    512,
    (65536, 65536, 65536),
    32,
    True,
    prefer_1d_tile=True,
    isCloseMemoryAsync=False,
)


@pointwise_dynamic(
    promotion_methods=[(0, 1, "ALWAYS_BOOL")],
    config=config_,
)
@triton.jit
def eq_func(x, y):
    return x.to(tl.float32) == y.to(tl.float32)


def eq(A, B):
    if A.device != B.device:
        if A.device.type == device:
            B = B.to(A.device)
        else:
            A = A.to(B.device)
    logger.debug("GEMS EQ")
    os.environ["TRITONXPU_COMPARE_FUSION"] = "1"
    os.environ["TRITONXPU_FP16_FAST"] = "1"
    res = eq_func(A, B)
    del os.environ["TRITONXPU_COMPARE_FUSION"]
    del os.environ["TRITONXPU_FP16_FAST"]
    return res


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "ALWAYS_BOOL")])
@triton.jit
def eq_func_scalar(x, y):
    return x.to(tl.float32) == y.to(tl.float32)


def eq_scalar(A, B):
    logger.debug("GEMS EQ SCALAR")
    return eq_func_scalar(A, B)



import torch
import torch.nn as nn

class RMSNorm(nn.Module):
    """nlp 领域"""
    def __init__(self, dim):
        """
        :param dim: 输入的维度
        """
        super(RMSNorm, self).__init__()
        self.scale = nn.Parameter(torch.ones(dim))  # 可学习的缩放参数
    
    def forward(self, x):
        # x 的形状为 [batch_size, seq_len, dim]
        
        # 计算均方根 (RMS) shape is [2, 4, 1]
        rms = torch.sqrt(torch.mean(x ** 2, dim=-1, keepdim=True))
        
        # 归一化，并应用缩放参数
        return x / rms * self.scale

# 测试 RMSNorm
batch_size, seq_len, dim = 2, 4, 8
x = torch.randn(batch_size, seq_len, dim)

rmsnorm = RMSNorm(dim)
output = rmsnorm(x)

# nn.RMSNorm 如果传入的是单个整数，则会将其视为一个单元素列表，
# 模块会对最后一个维度进行归一化，并且该维度的大小应该符合这个指定值。
rmsnorm_pytorch = nn.RMSNorm(dim)
output_pytorch = rmsnorm_pytorch(x)
    
# print("输入:\n", x)
# print("输出:\n", output)
print("输入和输出的形状: ", x.shape, output.shape)

if torch.allclose(output, output_pytorch, atol=1e-6):
    print("结果验证通过: 自己实的 RMSNorm 和 pytorch nn.RMSNorm 结果一致！")
else:
    print("结果验证失败: 自己实的 RMSNorm 和 pytorch nn.RMSNorm 结果不一致。")


import math
import torch
import triton
import triton.language as tl

def get_fwd_config(B, H, M, N, D, causal):
    return (32, 32, 1, 4)

def total_attention(q, k, l, causal=False, sm_scale=None):
    Dq, Dk = q.shape[-1], k.shape[-1]
    assert Dq == Dk
    assert Dk in {16, 32, 64, 128}
    # assert L is contiguous

    B, H, M, D = q.shape
    N = k.shape[2]
    Hk = k.shape[1]
    assert H % Hk == 0, "number of heads in q must be a multiple of that in k"
    num_groups = H // Hk

    P_SEQ = N - M

    if sm_scale is None:
        sm_scale = 1. / math.sqrt(D)

    # to work around https://github.com/openai/triton/issues/2441
    device = torch.cuda.device_of(q)
    with torch.cuda.device(device):
        config = get_fwd_config(B, H, M, N, D, causal)
        BLOCK_M, BLOCK_N, num_stages, num_warps = config

        divisible_m = M % BLOCK_M == 0
        divisible_n = N % BLOCK_N == 0
        # consider using 3d grid to avoid div & rem
        grid = (triton.cdiv(N, BLOCK_N), H, B)
        tot_attn = torch.empty((B, H, N), dtype=torch.float32, device=q.device)
        _total_attention_kernel[grid](
            q, k, l, tot_attn, sm_scale,
            q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            B, H, M, N, P_SEQ, num_groups,
            BLOCK_M=BLOCK_M, BLOCK_DMODEL=D, BLOCK_N=BLOCK_N,
            CAUSAL=causal,
            DIVISIBLE_M=divisible_m, DIVISIBLE_N=divisible_n,
            num_stages=num_stages, num_warps=num_warps,
        )
    return tot_attn


@triton.jit
def _total_attention_kernel(
    Q, K, L, TA, sm_scale,
    stride_qz, stride_qh, stride_qm, stride_qk,
    stride_kz, stride_kh, stride_kn, stride_kk,
    Z, H, M, N, P_SEQ, num_groups,
    BLOCK_M: tl.constexpr, BLOCK_DMODEL: tl.constexpr, BLOCK_N: tl.constexpr,
    CAUSAL: tl.constexpr,
    DIVISIBLE_M: tl.constexpr, DIVISIBLE_N: tl.constexpr,
):
    # -- grid id --
    start_n = tl.program_id(0)
    off_h = tl.program_id(1)
    off_z = tl.program_id(2)
    log2e: tl.constexpr = 1.4426950408889634
    qk_scale = sm_scale * log2e

    # offset pointers for (batch, head)
    off_hk = off_h // num_groups
    Q += off_z * stride_qz + off_h * stride_qh
    K += off_z * stride_kz + off_hk * stride_kh
    L += (off_z * H + off_h) * M
    TA += (off_z * H + off_h) * N # (b, h, n)

    if CAUSAL:
        lo = tl.maximum(start_n * BLOCK_N - P_SEQ, 0)
        lo = (lo // BLOCK_M) * BLOCK_M
    else:
        lo = 0

    offs_m_init = lo + tl.arange(0, BLOCK_M)
    offs_n = start_n * BLOCK_N + tl.arange(0, BLOCK_N)
    offs_m_base = tl.arange(0, BLOCK_M)
    offs_k = tl.arange(0, BLOCK_DMODEL)

    # initialize pointers to value-like data
    q_ptrs = Q + (offs_m_init[:, None] * stride_qm + offs_k[None, :] * stride_qk) # (BLOCK_M, BLOCK_DMODEL)
    k_ptrs = K + (offs_n[:, None] * stride_kn + offs_k[None, :] * stride_kk) # (BLOCK_N, BLOCK_DMODEL)
    ta_ptrs = TA + offs_n # (BLOCK_N, )

    # k and v stay in SRAM throughout
    if DIVISIBLE_N:
        k = tl.load(k_ptrs)
    else:
        mask_n = offs_n < N
        k = tl.load(k_ptrs, mask=mask_n[:, None])

    # initialize total attention
    tot_attn = tl.zeros([BLOCK_N], dtype=tl.float32)

    # loop over a col
    for start_m in range(lo, M, BLOCK_M):
        start_m = tl.multiple_of(start_m, BLOCK_M)
        offs_m = start_m + offs_m_base
        causal_mask = (P_SEQ + offs_m[:, None]) >= (offs_n[None, :]) # (BLOCK_M, BLOCK_N)

        if DIVISIBLE_M:
            q = tl.load(q_ptrs)
        else:
            mask_m = offs_m < M
            valid_mask = mask_m[:, None] # & mask_n
            q = tl.load(q_ptrs, mask=mask_m[:, None])
        # recompute p = softmax(qk * sm_scale, dim=-1)
        s = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        s += tl.dot(q, tl.trans(k))

        # NOTE: since softmax in backward is pointwise, the normalizer has been saved in fwd)
        # So masking on s is not needed.
        # s = tl.where(valid_mask, s , float("-inf"))
        # if CAUSAL:
        #     s = tl.where(causal_mask, s, float("-inf"))

        # -- recompute p ---
        if DIVISIBLE_M:
            l = tl.load(L + offs_m)
        else:
            l = tl.load(L + offs_m, mask=mask_m)
        p = tl.math.exp2(s * qk_scale - l[:, None] * log2e) # (BLOCK_M, BLOCK_N)

        if not DIVISIBLE_M:
            p = tl.where(valid_mask, p, 0.0)
        if CAUSAL:
            p = tl.where(causal_mask, p, 0.0)

        tot_attn += tl.sum(p, 0)
        # increment pointers
        q_ptrs += BLOCK_M * stride_qm


    if DIVISIBLE_N:
        tl.store(ta_ptrs, tot_attn) # (BLOCK_N,)
    else:
        tl.store(ta_ptrs, tot_attn, mask=mask_n) # (BLOCK_N, )





import torch


def naive_recurrent_rwkv6(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    w: torch.Tensor,
    u: torch.Tensor,
    scale: float | None = None,
    initial_state: torch.Tensor | None = None,
    output_final_state: bool | None = False,
):
    orig_dtype = q.dtype
    B, H, T, K, V = *q.shape, v.shape[-1]
    q, k, v, w, u = map(lambda x: x.float(), (q, k, v, w, u))
    h = torch.zeros(B, H, K, V, dtype=torch.float32, device=q.device)
    o = torch.zeros_like(v)

    if scale is None:
        scale = K ** -0.5

    if initial_state is not None:
        h += initial_state

    for i in range(T):
        q_i = q[:, :, i, :] * scale
        k_i = k[:, :, i]
        v_i = v[:, :, i, :]
        w_i = w[:, :, i].exp()
        kv_i = k_i[..., None] * v_i[..., None, :]
        o_i = (h + u[None, ..., None] * kv_i) * q_i[..., None]
        o[:, :, i] = o_i.sum(-2)
        h = h * w_i[..., None] + kv_i
    ht = h if output_final_state else None
    return o.to(orig_dtype), ht


@torch.no_grad
@torch.jit.script
def naive_recurrent_rwkv6_bwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    w: torch.Tensor,
    u: torch.Tensor,
    o: torch.Tensor,
    do: torch.Tensor,
    initial_state: torch.Tensor | None = None,
):
    q, k, v, w, u, o, do = (x.to(dtype=torch.float32) for x in (q, k, v, w, u, o, do))
    B, H, T, K, V = q.shape[0], q.shape[1], q.shape[2], q.shape[3], v.shape[-1]
    h = torch.zeros(B, H, K, V, dtype=torch.float32, device=q.device)
    dq = torch.zeros_like(q)
    dq_aux = torch.zeros_like(q)

    if initial_state is not None:
        h += initial_state

    for i in range(T):
        k_i = k[:, :, i]
        v_i = v[:, :, i]
        w_i = w[:, :, i].exp()
        kv_i = k_i[..., None] * v_i[..., None, :]
        h_i = (h + u[None, ..., None] * kv_i)
        dq_i = (do[:, :, i, None, :] * h_i).sum(-1)
        dq_aux_i = (do[:, :, i, None, :] * h).sum(-1)
        dq[:, :, i] = dq_i
        dq_aux[:, :, i] = dq_aux_i
        h = h * w_i[..., None] + kv_i

    du = torch.zeros_like(u)
    dh = torch.zeros_like(h)
    dk = torch.zeros_like(k)
    dk_aux = torch.zeros_like(k)
    dv = torch.zeros_like(v)

    for i in range(T - 1, -1, -1):
        d_kv_i = do[:, :, i, None, :] * q[:, :, i, :, None]
        k_i = k[:, :, i]
        v_i = v[:, :, i]
        du_i = (d_kv_i * k_i[..., None] * v_i[..., None, :]).sum(-1)
        du += du_i.sum(0)
        dk_i = (dh * v_i[..., None, :]).sum(-1)
        dk_aux[:, :, i] = dk_i
        dk_i += (d_kv_i * u[None, ..., None] * v_i[..., None, :]).sum(-1)
        dv_i = (d_kv_i * u[None, ..., None] * k_i[..., None]).sum(-2)
        dv_i += (dh * k_i[..., None]).sum(-2)

        dk[:, :, i] = dk_i
        dv[:, :, i] = dv_i
        dh = dh * w[:, :, i, :, None].exp() + d_kv_i

    # dw = q * dq_aux - k * dk_aux
    dw = torch.zeros_like(w)
    for i in range(T - 2, -1, -1):
        dw[:, :, i] = dw[:, :, i+1] + dq_aux[:, :, i+1] * q[:, :, i+1] - dk_aux[:, :, i] * k[:, :, i]

    return dq, dk, dv, dw, du, dh



"""
NeuroBrix Validators - Synchronization Validator.

Validates cross-device synchronization and data dependencies.

ZERO FALLBACK: Sync issues raise explicit SyncError.
"""

from typing import Any, Dict, List, Set, Optional, Tuple

from .base import BaseValidator, SafetyLevel, ValidationResult, SyncError, DeadlockRisk
from .config import get_config


class SyncValidator(BaseValidator):
    """
    Validates synchronization correctness.

    Checks:
    - Cross-device data dependencies
    - Stream ordering
    - Event signaling
    - Deadlock detection
    - Memory fence requirements
    """

    @property
    def name(self) -> str:
        return "SyncValidator"

    def __init__(self, safety_level: SafetyLevel = SafetyLevel.STANDARD):
        super().__init__(safety_level)
        self.config = get_config()

    def validate(self, graph: Any, context: Dict[str, Any]) -> ValidationResult:
        """
        Validate synchronization in the graph.

        Args:
            graph: Execution graph with nodes
            context: Must contain:
                - 'device_assignments': Dict mapping node -> device
                - 'stream_assignments': Dict mapping node -> stream (optional)

        Returns:
            ValidationResult with sync validation status
        """
        warnings = []
        errors = []
        metrics = {}

        device_assignments = context.get('device_assignments', {})
        stream_assignments = context.get('stream_assignments', {})

        if not device_assignments:
            return ValidationResult(
                passed=True,
                validator_name=self.name,
                message="No device assignments to validate",
                warnings=warnings,
            )

        # Analyze device topology
        devices = set(device_assignments.values())
        metrics['num_devices'] = len(devices)
        metrics['devices'] = list(devices)

        # Check cross-device transfers
        transfers = self._find_cross_device_transfers(graph, device_assignments)
        metrics['cross_device_transfers'] = len(transfers)

        for transfer in transfers:
            self._validate_transfer(transfer, context, warnings, errors)

        # Check for potential deadlocks
        if self.safety_level >= SafetyLevel.STANDARD:
            deadlock_risks = self._detect_deadlock_risks(
                graph, device_assignments, stream_assignments
            )
            for risk in deadlock_risks:
                warnings.append(f"Potential deadlock risk: {risk}")
            metrics['deadlock_risks'] = len(deadlock_risks)

        # Check stream depth
        if stream_assignments:
            self._check_stream_depth(stream_assignments, warnings, errors, metrics)

        if errors:
            return ValidationResult(
                passed=False,
                validator_name=self.name,
                message="\n".join(errors),
                warnings=warnings,
                metrics=metrics,
            )

        return ValidationResult(
            passed=True,
            validator_name=self.name,
            message=f"Sync validated: {len(devices)} devices, {len(transfers)} transfers",
            warnings=warnings,
            metrics=metrics,
        )

    def _find_cross_device_transfers(
        self,
        graph: Any,
        device_assignments: Dict[str, str],
    ) -> List[Dict[str, Any]]:
        """Find all cross-device data transfers."""
        transfers = []

        nodes = getattr(graph, 'nodes', [])
        if hasattr(graph, '__iter__') and not nodes:
            nodes = list(graph)

        for node in nodes:
            node_name = getattr(node, 'name', str(node))
            node_device = device_assignments.get(node_name)

            if node_device is None:
                continue

            # Check inputs
            inputs = getattr(node, 'inputs', [])
            for inp in inputs:
                inp_name = getattr(inp, 'name', str(inp))
                inp_device = device_assignments.get(inp_name)

                if inp_device is not None and inp_device != node_device:
                    transfers.append({
                        'from_node': inp_name,
                        'from_device': inp_device,
                        'to_node': node_name,
                        'to_device': node_device,
                    })

        return transfers

    def _validate_transfer(
        self,
        transfer: Dict[str, Any],
        context: Dict[str, Any],
        warnings: List[str],
        errors: List[str],
    ) -> None:
        """Validate a single cross-device transfer."""
        from_device = transfer['from_device']
        to_device = transfer['to_device']

        # Check if devices can communicate
        valid_pairs = context.get('valid_device_pairs', None)
        if valid_pairs is not None:
            pair = (from_device, to_device)
            if pair not in valid_pairs and (to_device, from_device) not in valid_pairs:
                errors.append(
                    f"Invalid device transfer: {from_device} -> {to_device}"
                )

        # Check for GPU -> CPU -> GPU pattern (inefficient)
        if 'cpu' in to_device.lower() and 'gpu' in from_device.lower():
            warnings.append(
                f"GPU -> CPU transfer detected: {transfer['from_node']} -> {transfer['to_node']}. "
                f"Consider keeping data on GPU if possible."
            )

    def _detect_deadlock_risks(
        self,
        graph: Any,
        device_assignments: Dict[str, str],
        stream_assignments: Dict[str, str],
    ) -> List[str]:
        """Detect potential deadlock situations."""
        risks = []

        # Build dependency graph per device
        device_deps: Dict[str, List[Tuple[str, str]]] = {}

        nodes = getattr(graph, 'nodes', [])
        if hasattr(graph, '__iter__') and not nodes:
            nodes = list(graph)

        for node in nodes:
            node_name = getattr(node, 'name', str(node))
            node_device = device_assignments.get(node_name)

            if node_device is None:
                continue

            if node_device not in device_deps:
                device_deps[node_device] = []

            inputs = getattr(node, 'inputs', [])
            for inp in inputs:
                inp_name = getattr(inp, 'name', str(inp))
                inp_device = device_assignments.get(inp_name)

                if inp_device is not None and inp_device != node_device:
                    device_deps[node_device].append((inp_device, node_name))

        # Check for circular dependencies between devices
        visited: Set[str] = set()
        rec_stack: Set[str] = set()

        def has_cycle(device: str) -> bool:
            visited.add(device)
            rec_stack.add(device)

            deps = device_deps.get(device, [])
            for dep_device, _ in deps:
                if dep_device not in visited:
                    if has_cycle(dep_device):
                        return True
                elif dep_device in rec_stack:
                    return True

            rec_stack.remove(device)
            return False

        for device in device_deps:
            if device not in visited:
                if has_cycle(device):
                    risks.append(f"Circular dependency involving device {device}")

        # Check stream conflicts
        if stream_assignments:
            stream_nodes: Dict[str, List[str]] = {}
            for node_name, stream in stream_assignments.items():
                if stream not in stream_nodes:
                    stream_nodes[stream] = []
                stream_nodes[stream].append(node_name)

            # Check for same stream with cross-device ops
            for stream, nodes_in_stream in stream_nodes.items():
                devices_in_stream = set()
                for node_name in nodes_in_stream:
                    device = device_assignments.get(node_name)
                    if device:
                        devices_in_stream.add(device)

                if len(devices_in_stream) > 1:
                    risks.append(
                        f"Stream {stream} contains nodes on multiple devices: "
                        f"{devices_in_stream}"
                    )

        return risks

    def _check_stream_depth(
        self,
        stream_assignments: Dict[str, str],
        warnings: List[str],
        errors: List[str],
        metrics: Dict[str, Any],
    ) -> None:
        """Check stream depth limits."""
        config = self.config

        # Count unique streams
        unique_streams = set(stream_assignments.values())
        num_streams = len(unique_streams)
        metrics['num_streams'] = num_streams

        if num_streams > config.max_stream_depth:
            errors.append(
                f"Stream depth {num_streams} exceeds max {config.max_stream_depth}"
            )
        elif num_streams > config.max_stream_depth * 0.8:
            warnings.append(
                f"High stream count: {num_streams}/{config.max_stream_depth}"
            )

    def validate_device_transfer(
        self,
        from_device: str,
        to_device: str,
        size_bytes: int,
    ) -> None:
        """
        Validate a single device transfer.

        Raises SyncError if transfer is invalid.
        """
        if from_device == to_device:
            return  # Same device, no transfer needed

        # Check for incompatible transfers
        incompatible = [
            ('cpu', 'disk'),
            ('disk', 'cpu'),
        ]

        pair = (from_device.lower(), to_device.lower())
        if pair in incompatible:
            raise SyncError(
                f"Incompatible device transfer: {from_device} -> {to_device}",
                device_a=from_device,
                device_b=to_device,
                suggestion="Use intermediate staging buffer",
            )

    def check_for_deadlock(
        self,
        dependencies: Dict[str, List[str]],
    ) -> Optional[List[str]]:
        """
        Check for deadlock in dependency graph.

        Args:
            dependencies: Dict mapping node -> list of dependencies

        Returns:
            Cycle path if deadlock found, None otherwise
        """
        visited: Set[str] = set()
        rec_stack: Set[str] = set()
        path: List[str] = []

        def dfs(node: str) -> Optional[List[str]]:
            visited.add(node)
            rec_stack.add(node)
            path.append(node)

            for dep in dependencies.get(node, []):
                if dep not in visited:
                    result = dfs(dep)
                    if result is not None:
                        return result
                elif dep in rec_stack:
                    # Found cycle
                    cycle_start = path.index(dep)
                    return path[cycle_start:] + [dep]

            path.pop()
            rec_stack.remove(node)
            return None

        for node in dependencies:
            if node not in visited:
                cycle = dfs(node)
                if cycle is not None:
                    return cycle

        return None

    def should_run(self, safety_level: SafetyLevel) -> bool:
        """Sync validation runs at STANDARD and PARANOID levels."""
        return safety_level >= SafetyLevel.STANDARD



"""
InferenceEngine — Persistent model serving for NeuroBrix.

Load once, serve many. Weights stay in VRAM between requests.
Works for ALL families: LLM, image, audio, video.

Strategy-aware:
  Warm strategies (single_gpu eager, pipeline, tp, fgp):
    Weights persist in VRAM across requests. Near-zero latency after first request.
  Cold strategies (lazy, zero3, pp_lazy):
    Weights reload per request (model too large for VRAM). Setup + context still persist.

ZERO SEMANTIC: Family-aware only for output formatting, not for execution.
ZERO HARDCODE: All config from NBX container + Prism hardware profile.
"""

import time
import json
import torch
from pathlib import Path
from typing import Any, Dict, Optional

from neurobrix.serving.session import ConversationSession

# Warm serving is now determined by loading_mode from Prism solver.
# loading_mode="eager" → weights stay in VRAM (warm serving)
# loading_mode="lazy"  → weights reload per request (cold serving)
# No manual strategy set needed — Prism is the single source of truth.


class InferenceEngine:
    """
    Persistent inference engine — load once, serve many.

    Wraps RuntimeExecutor and keeps it alive across requests.
    For warm strategies: GraphExecutor weights persist in VRAM via _persistent flag.
    For cold strategies: weights reload per request but setup is reused.

    For LLM models: manages ConversationSession for multi-turn chat.
    For image/audio/video: each generate() call is independent.
    """

    def __init__(self, model_name: str, hardware_id: str, mode: str = "compiled"):
        self.model_name = model_name
        self.hardware_id = hardware_id
        self.mode = mode

        self._pkg = None           # RuntimePackage (immutable)
        self._plan = None          # ExecutionPlan (immutable)
        self._executor = None      # RuntimeExecutor (persistent, setup once)
        self._container = None     # NBXContainer
        self._family = None        # "llm", "image", "audio", "video"
        self._session = None       # ConversationSession (LLM only)
        self._is_loaded = False
        self._warm_serving = False  # True if strategy supports persistent weights

    def load(self) -> None:
        """
        One-time setup: container -> Prism -> RuntimeExecutor.setup().

        After this call, the engine is ready to serve requests.
        First request loads weights into VRAM. For warm strategies,
        weights persist across all subsequent requests.
        """
        from neurobrix.nbx import NBXContainer
        from neurobrix.core.prism import PrismSolver, load_profile, InputConfig
        from neurobrix.core.runtime.loader import NBXRuntimeLoader
        from neurobrix.core.runtime.executor import RuntimeExecutor
        from neurobrix.cli.utils import find_model

        t_start = time.time()

        # 1. Find model
        nbx_path = find_model(self.model_name)

        # 2. Load container
        self._container = NBXContainer.load(str(nbx_path))
        manifest = self._container.get_manifest() or {}
        self._family = manifest.get("family")
        neural_components = self._container.get_neural_components()

        # 3. Solve Prism allocation
        hw_profile = load_profile(self.hardware_id)

        cache_path = self._container._cache_path
        defaults_path = cache_path / "runtime" / "defaults.json"
        cached_defaults = {}
        if defaults_path.exists():
            with open(defaults_path) as f:
                cached_defaults = json.load(f)

        height = cached_defaults.get("height", 1024)
        width = cached_defaults.get("width", 1024)
        vae_scale = cached_defaults.get("vae_scale_factor", 8)

        input_config = InputConfig(
            batch_size=2,
            height=height,
            width=width,
            dtype="float16",
            vae_scale=vae_scale,
        )

        solver = PrismSolver()
        self._plan = solver.solve_smart(self._container, hw_profile, input_config, serve_mode=True)

        # 4. Determine warm serving compatibility
        # loading_mode is set by Prism solver based on strategy type:
        #   eager = weights stay in VRAM (single_gpu, pp, fgp, tp)
        #   lazy  = weights reload per request (pp_lazy, lazy_sequential, zero3)
        loading_mode = getattr(self._plan, 'loading_mode', 'lazy')
        self._warm_serving = (loading_mode == "eager")

        # 5. Load RuntimePackage
        loader = NBXRuntimeLoader()
        self._pkg = loader.load(str(nbx_path))

        # 6. Create RuntimeExecutor and run one-time setup
        self._executor = RuntimeExecutor(self._pkg, self._plan, mode=self.mode)

        # Set persistent mode BEFORE setup/execute so GraphExecutors get the flag
        if self._warm_serving:
            self._executor._persistent_mode = True

        self._executor.setup()

        # 7. Pre-warm weights for warm strategies (load into VRAM now, not on R1)
        if self._warm_serving:
            warmup_inputs = {"global.prompt": "warmup", "global.max_tokens": 1}
            if self._family == "llm":
                warmup_inputs["global.chat_mode"] = True
            self._executor.execute(warmup_inputs)

        t_load = time.time() - t_start
        self._is_loaded = True

        print(f"[NeuroBrix] {self.model_name} | {self.hardware_id} | {self._plan.strategy} | Ready in {t_load:.2f}s")

        if not self._warm_serving:
            print(f"[NeuroBrix] Cold serving — weights reload per request, setup/context persists")

    def generate(self, prompt: str, **kwargs) -> Dict[str, Any]:
        """
        Single-shot generation. Reused executor, fresh variable resolution.

        Works for ALL families. Returns structured result dict.
        """
        if not self._is_loaded:
            raise RuntimeError("ZERO FALLBACK: Engine not loaded. Call load() first.")

        # Build inputs dict (same structure as cmd_run)
        inputs = {"global.prompt": prompt}
        if "steps" in kwargs and kwargs["steps"] is not None:
            inputs["global.num_inference_steps"] = kwargs["steps"]
        if "height" in kwargs and kwargs["height"] is not None:
            inputs["global.height"] = kwargs["height"]
        if "width" in kwargs and kwargs["width"] is not None:
            inputs["global.width"] = kwargs["width"]
        if "cfg" in kwargs and kwargs["cfg"] is not None:
            inputs["global.guidance_scale"] = kwargs["cfg"]
        if "temperature" in kwargs and kwargs["temperature"] is not None:
            inputs["global.temperature"] = kwargs["temperature"]
        if "repetition_penalty" in kwargs and kwargs["repetition_penalty"] is not None:
            inputs["global.repetition_penalty"] = kwargs["repetition_penalty"]
        if "max_tokens" in kwargs and kwargs["max_tokens"] is not None:
            inputs["global.max_tokens"] = kwargs["max_tokens"]
        if "chat_mode" in kwargs:
            inputs["global.chat_mode"] = kwargs["chat_mode"]
        if "seed" in kwargs and kwargs["seed"] is not None:
            seed = kwargs["seed"]
            inputs["global.seed"] = seed
            torch.manual_seed(seed)
            if torch.cuda.is_available():
                torch.cuda.manual_seed_all(seed)

        # Execute
        t_start = time.time()
        if torch.cuda.is_available():
            torch.cuda.synchronize()

        outputs = self._executor.execute(inputs)

        if torch.cuda.is_available():
            torch.cuda.synchronize()
        t_total = time.time() - t_start

        # Build result
        result = {"timing": {"total_s": round(t_total, 3)}, "family": self._family}

        if self._family == "llm":
            result.update(self._extract_llm_output(outputs))
        else:
            result["outputs"] = outputs

        return result

    def _generate_from_token_ids(self, token_ids: list, **kwargs) -> Dict[str, Any]:
        """
        Generate from pre-tokenized token IDs (chat path).

        Bypasses TextProcessor entirely — token IDs already include chat template,
        special tokens, and generation prompt. No double-BOS possible.
        """
        if not self._is_loaded:
            raise RuntimeError("ZERO FALLBACK: Engine not loaded. Call load() first.")

        # Build inputs with pre-tokenized IDs instead of string prompt
        inputs = {
            "global.input_token_ids": token_ids,
            "global.chat_mode": False,  # Template already applied
        }
        if "temperature" in kwargs and kwargs["temperature"] is not None:
            inputs["global.temperature"] = kwargs["temperature"]
        if "repetition_penalty" in kwargs and kwargs["repetition_penalty"] is not None:
            inputs["global.repetition_penalty"] = kwargs["repetition_penalty"]
        if "max_tokens" in kwargs and kwargs["max_tokens"] is not None:
            inputs["global.max_tokens"] = kwargs["max_tokens"]
        if "seed" in kwargs and kwargs["seed"] is not None:
            seed = kwargs["seed"]
            inputs["global.seed"] = seed
            torch.manual_seed(seed)
            if torch.cuda.is_available():
                torch.cuda.manual_seed_all(seed)

        # Execute
        t_start = time.time()
        if torch.cuda.is_available():
            torch.cuda.synchronize()

        outputs = self._executor.execute(inputs)

        if torch.cuda.is_available():
            torch.cuda.synchronize()
        t_total = time.time() - t_start

        result = {"timing": {"total_s": round(t_total, 3)}, "family": self._family}
        result.update(self._extract_llm_output(outputs))
        return result

    def chat(self, message: str, **kwargs) -> str:
        """
        Multi-turn chat for LLM models. Manages conversation session.

        Builds full context from conversation history each turn.
        Context overflow protection: summarizes old turns if KV cache limit approached.
        """
        if self._family != "llm":
            raise RuntimeError(
                f"ZERO FALLBACK: chat() requires LLM family, got '{self._family}'."
            )

        if self._session is None:
            tokenizer = self._executor.modules.get("tokenizer")
            if tokenizer is None:
                raise RuntimeError("ZERO FALLBACK: No tokenizer module in executor.")
            self._session = ConversationSession(
                tokenizer=tokenizer,
                defaults=self._pkg.defaults,
            )

        # Add user message to history
        self._session.add_user_message(message)

        # Context overflow protection
        max_cache_len = self._get_max_cache_len()
        max_tokens = kwargs.get("max_tokens") or self._pkg.defaults.get("max_tokens", 512)
        summarized = self._session.ensure_fits(
            max_cache_len=max_cache_len,
            max_tokens=max_tokens,
            generate_fn=self.generate,
        )
        if summarized:
            print("[Engine] Conversation history compressed — older turns summarized")

        # Build pre-tokenized IDs from conversation history via chat_template.
        # This avoids double-BOS: the template already includes special tokens,
        # so we pass token IDs directly to the executor, bypassing TextProcessor.
        token_ids = self._session.build_token_ids()
        result = self._generate_from_token_ids(token_ids, **kwargs)

        # Extract response text
        response_text = result.get("text", "")
        self._session.add_assistant_message(response_text)

        return response_text

    def _get_max_cache_len(self) -> int:
        """Get KV cache capacity from Prism plan."""
        kv_plan = getattr(self._plan, 'kv_cache_plan', None)
        if kv_plan is not None:
            return kv_plan.max_cache_len
        # Fallback to max_position_embeddings from lm_config
        lm_config = self._pkg.defaults.get("lm_config", {})
        return lm_config.get("max_position_embeddings", 2048)

    def new_conversation(self) -> None:
        """Start a new conversation, clearing history."""
        if self._session is not None:
            self._session.clear()

    def unload(self) -> None:
        """Release all VRAM. Engine can be re-loaded later."""
        # Clear persistent flags so weights can actually be freed
        if self._executor is not None:
            for _, executor in self._executor.executors.items():
                if hasattr(executor, '_persistent'):
                    executor._persistent = False

            for name in list(self._executor.executors.keys()):
                self._executor._unload_component_weights(name)

        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        self._is_loaded = False

    def get_status(self) -> Dict[str, Any]:
        """Return engine status info."""
        status = {
            "model": self.model_name,
            "hardware": self.hardware_id,
            "family": self._family,
            "loaded": self._is_loaded,
            "mode": self.mode,
            "warm_serving": self._warm_serving,
        }

        if self._plan is not None:
            status["strategy"] = self._plan.strategy

        if self._is_loaded and torch.cuda.is_available():
            vram_used = 0
            for i in range(torch.cuda.device_count()):
                vram_used += torch.cuda.memory_allocated(i)
            status["vram_used_gb"] = round(vram_used / 1024**3, 2)

        if self._session is not None:
            status["context"] = self._session.get_context_info()

        return status

    @property
    def is_loaded(self) -> bool:
        return self._is_loaded

    @property
    def family(self) -> Optional[str]:
        return self._family

    def save_output(self, outputs: Dict[str, Any], output_path: str) -> str:
        """
        Save non-LLM output (image/audio/video) to file. Returns saved path.
        Replicates cold-path logic from run.py.
        """
        import numpy as np
        from PIL import Image
        from neurobrix.core.config import get_output_processing
        from neurobrix.core.module.output_processor import OutputProcessor

        final_output = self._executor.get_final_output(outputs)
        if final_output is None:
            raise RuntimeError("ZERO FALLBACK: No tensor output from model")

        output_range = self._pkg.defaults.get("output_range")
        if output_range is None:
            output_cfg = get_output_processing(self._family)
            output_range = output_cfg.get("output_range", [-1.0, 1.0])

        output_cfg = get_output_processing(self._family)
        batch_axis = output_cfg.get("batch_axis", 0)
        layout = output_cfg.get("layout", "CHW")
        valid_channels = output_cfg.get("valid_channels", [1, 3, 4])
        channel_axis = output_cfg.get("channel_axis", 1)
        bit_depth = output_cfg.get("bit_depth", 8)

        img = torch.select(final_output, batch_axis, 0).cpu().float()

        processor = OutputProcessor.from_package(self._pkg)
        img = processor.process(img, output_range)
        img = img.clamp(0, 1)

        if layout == "CHW" and img.dim() == 3:
            actual_channel_axis = channel_axis - 1 if batch_axis < channel_axis else channel_axis
            if img.shape[actual_channel_axis] in valid_channels:
                img = img.permute(1, 2, 0)

        if bit_depth == 8:
            img_np = (img.numpy() * 255).astype(np.uint8)
        elif bit_depth == 16:
            img_np = (img.numpy() * 65535).astype(np.uint16)
        else:
            img_np = (img.numpy() * 255).astype(np.uint8)

        if img_np.shape[-1] == 1:
            img_np = img_np.squeeze(-1)

        Image.fromarray(img_np).save(output_path)
        return output_path

    def _extract_llm_output(self, outputs: Dict[str, Any]) -> Dict[str, Any]:
        """Extract text from LLM generation outputs."""
        output_tokens = outputs.get("output_tokens")
        if output_tokens is None:
            output_tokens = outputs.get("global.output_tokens")
        if output_tokens is None:
            return {"text": "", "tokens": 0}

        token_ids = output_tokens.flatten().tolist()
        text = ""

        tokenizer = self._executor.modules.get("tokenizer")
        if tokenizer is not None and hasattr(tokenizer, 'decode'):
            text = tokenizer.decode(token_ids, skip_special_tokens=True)

        return {
            "text": text,
            "tokens": len(token_ids),
        }



import logging
from typing import Optional, Tuple

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import device, torch_device_fn

device = device.name
logger = logging.getLogger(__name__)


@triton.autotune(
    configs=runtime.get_tuned_config("upsample_nearest2d"), key=["N", "C", "OH", "OW"]
)
@triton.heuristics(runtime.get_heuristic_config("upsample_nearest2d"))
@triton.jit
def upsample_nearest2d_kernel(
    ptr_o,
    ptr_i,
    N,
    C,
    OH,
    OW,
    IH,
    IW,
    reciprocal_scale_h,
    reciprocal_scale_w,
    BLOCK_SIZE: tl.constexpr,
    SAME_H: tl.constexpr,
    SAME_W: tl.constexpr,
    USE_INT32_IDX: tl.constexpr,
):
    if USE_INT32_IDX:
        pid = tl.program_id(axis=0)
    else:
        pid = tl.program_id(axis=0).to(tl.int64)
    nc_stride = tl.num_programs(axis=1)
    NC = N * C
    nc_iter = tl.program_id(axis=1)
    pid = tl.program_id(axis=0)
    idx = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    ow = idx % OW
    oh = idx // OW % OH
    if SAME_H:
        ih = oh
    else:
        # tl.floor() cannot be found in 2.3.1, using int trunc
        ih = tl.minimum((oh * reciprocal_scale_h).to(tl.int32), IH - 1)
    if SAME_W:
        iw = ow
    else:
        iw = tl.minimum((ow * reciprocal_scale_w).to(tl.int32), IW - 1)

    offset_o = (nc_iter * OH + oh) * OW + ow
    offset_i = (nc_iter * IH + ih) * IW + iw
    src_index_stride = nc_stride * IH * IW
    dst_index_stride = nc_stride * OH * OW
    while nc_iter < NC:
        data = tl.load(ptr_i + offset_i)
        tl.store(ptr_o + offset_o, data)
        ptr_i += src_index_stride
        ptr_o += dst_index_stride
        nc_iter += nc_stride


def upsample_nearest2d(
    input: torch.Tensor,
    output_size: Tuple[int],
    scales_h: Optional[float] = None,
    scales_w: Optional[float] = None,
) -> torch.Tensor:
    logger.debug("GEMS UPSAMPLE NEAREST2D")
    assert input.device.type == device
    assert input.ndim == 4, "The ndim of input must be 4"
    assert len(output_size) == 2, "The len of output_size must be 2"
    OH, OW = output_size
    N, C, IH, IW = input.shape
    if scales_h is not None:
        reciprocal_scale_h = 1 / scales_h
    else:
        reciprocal_scale_h = IH / OH
    if scales_w is not None:
        reciprocal_scale_w = 1 / scales_w
    else:
        reciprocal_scale_w = IW / OW
    # allocate output
    output = torch.empty((N, C, OH, OW), device=input.device, dtype=input.dtype)
    total_threads = OH * OW
    grid = lambda META: (
        triton.cdiv(total_threads, META["BLOCK_SIZE"]),
        triton.cdiv(N * C, 4),
    )

    with torch_device_fn.device(input.device):
        upsample_nearest2d_kernel[grid](
            output,
            input,
            N,
            C,
            OH,
            OW,
            IH,
            IW,
            reciprocal_scale_h,
            reciprocal_scale_w,
        )
    return output



import logging

import torch
import triton
from triton import language as tl

from flag_gems.ops.mv import mv
from flag_gems.utils import libentry

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


# The outer kernel requires 3 parameters to determine the splitting method,
# but during actual tuning, you only need to determine the total size of the split blocks.
# Based on the second input length N and the total size of the split blocks,
# the 3 parameters that determine the splitting method can be calculated.
# Therefore, the conversion between these two is achieved through early_config_prune.
def early_config_prune(configs, named_args, **kwargs):
    if "N" in kwargs:
        N = kwargs["N"]
    else:
        N = named_args["N"]

    new_configs = []
    for config in configs:
        tile_size = config.kwargs["tile_size"]
        block_n = min(tile_size, N)
        block_m = triton.cdiv(tile_size, block_n)
        new_config = triton.Config(
            {"BLOCK_M": block_m, "BLOCK_N": block_n, "NEED_LOOP_N": block_n < N},
            num_stages=config.num_stages,
            num_warps=config.num_warps,
        )
        new_configs.append(new_config)

    return new_configs


@libentry()
@triton.autotune(
    configs=[
        triton.Config({"tile_size": 1024}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 2048}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 4096}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 8192}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 16384}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 21760}, num_stages=3, num_warps=1),
        triton.Config({"tile_size": 32768}, num_stages=3, num_warps=1),
    ],
    key=["M", "N"],
    prune_configs_by={"early_config_prune": early_config_prune},
)
@triton.jit
def outer_kernel(
    lhs,
    rhs,
    res,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    NEED_LOOP_N: tl.constexpr,
):
    pid = tl.program_id(0)
    num_jobs = tl.num_programs(axis=0)

    m_tasks_num = tl.cdiv(M, BLOCK_M)
    n_tasks_num = tl.cdiv(N, BLOCK_N)
    total_tasks_num = m_tasks_num * n_tasks_num

    if NEED_LOOP_N:
        for task_id in range(pid, total_tasks_num, num_jobs):
            start_m = task_id // n_tasks_num
            start_n = task_id % n_tasks_num

            offset_m = tl.arange(0, BLOCK_M) + start_m * BLOCK_M
            lhs_val = tl.load(lhs + offset_m, mask=offset_m < M)

            offset_n = tl.arange(0, BLOCK_N) + start_n * BLOCK_N
            rhs_val = tl.load(rhs + offset_n, mask=offset_n < N)

            res_val = lhs_val[:, None] * rhs_val[None, :]

            offset_r = offset_m[:, None] * N + offset_n[None, :]
            tl.store(
                res + offset_r,
                res_val,
                mask=(offset_m[:, None] < M) & (offset_n[None, :] < N),
            )
    else:
        offset_n = tl.arange(0, BLOCK_N)
        rhs_val = tl.load(rhs + offset_n)
        for task_id in range(pid, total_tasks_num, num_jobs):
            start_m = task_id // n_tasks_num

            offset_m = tl.arange(0, BLOCK_M) + start_m * BLOCK_M
            lhs_val = tl.load(lhs + offset_m, mask=offset_m < M)

            res_val = lhs_val[:, None] * rhs_val[None, :]

            offset_r = offset_m[:, None] * N + offset_n[None, :]
            tl.store(
                res + offset_r,
                res_val,
                mask=(offset_m[:, None] < M) & (offset_n[None, :] < N),
            )


def outer_(lhs, rhs):
    m = lhs.shape[0]
    n = rhs.shape[0]
    res_shape = [m, n]
    res = torch.empty(res_shape, dtype=lhs.dtype, device="npu")
    grid = lambda META: (
        min(
            triton.cdiv(m, META["BLOCK_M"]) * triton.cdiv(n, META["BLOCK_N"]),
            65535,
        ),
    )
    outer_kernel[grid](lhs, rhs, res, m, n)
    return res


class Outer(torch.autograd.Function):
    @staticmethod
    def forward(ctx, inp, weight):
        logger.debug("GEMS_ASCEND OUTER")
        assert inp.ndim == 1 and weight.ndim == 1, "Invalid input"
        out = outer_(inp, weight)
        ctx.save_for_backward(inp, weight)
        return out

    @staticmethod
    def backward(ctx, out_grad):
        logger.debug("GEMS_ASCEND OUTER VJP")
        assert out_grad.ndim == 2, "invalide out_grad shape"

        inp, weight = ctx.saved_tensors

        inp_grad = mv(out_grad, weight)
        weight_grad = mv(out_grad.t(), inp)

        return inp_grad, weight_grad


def outer(inp, weight):
    return Outer.apply(inp, weight)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.utils import libentry, libtuner

from ..utils import MAX_NRAM_SIZE, TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


def get_max_block_size(dtype_size):
    return MAX_NRAM_SIZE // 3 // dtype_size


def config_prune(configs, named_args, **kwargs):
    N = named_args["N"]
    dtype_size = named_args["dtype_size"]
    max_block_size = get_max_block_size(dtype_size)

    pruned_configs = []
    index_block_size = []
    for config in configs:
        bs = config.kwargs["BLOCK_SIZE"]
        ibs = (bs + N - 1) // N
        if ibs not in index_block_size and ibs * N <= max_block_size:
            index_block_size.append(ibs)
            pruned_configs.append(config)

    in_n_elements = named_args["in_n_elements"]

    # make sure at least one config is at the load-balance sweet point
    if in_n_elements % TOTAL_CORE_NUM == 0:
        bs = min(max(in_n_elements // TOTAL_CORE_NUM, 1) * N, max_block_size)
    else:
        bs = min(max(in_n_elements // TOTAL_CORE_NUM, 1) * N + 1, max_block_size)
    if (bs + N - 1) // N not in index_block_size:
        pruned_configs.append(
            triton.Config(kwargs={"BLOCK_SIZE": bs}, num_stages=1, num_warps=1)
        )

    return pruned_configs


@triton.jit
def ld_st_1(indices, N: tl.constexpr, weight_ptr, in_mask, in_offsets, out_ptr):
    weight_offsets = indices[:, None] * N + tl.arange(0, N)
    embedding_weight = tl.load(weight_ptr + weight_offsets, in_mask[:, None])
    out_offsets = in_offsets[:, None] * N + tl.arange(0, N)
    tl.store(out_ptr + out_offsets, embedding_weight, in_mask[:, None])


@libentry()
@libtuner(
    configs=[
        # [512, 65536]
        triton.Config(kwargs={"BLOCK_SIZE": 512 * 2**i}, num_stages=1, num_warps=1)
        for i in range(0, 8, 2)
    ],
    key=["N"],
    prune_configs_by={
        "early_config_prune": config_prune,
    },
)
@triton.jit
def one_batch_index_select_kernel(  # 2D
    out_ptr,
    in_ptr,
    in_n_elements,
    weight_ptr,
    N: tl.constexpr,
    dtype_size,
    inp_numel,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(0)
    num_jobs = tl.num_programs(axis=0)

    INDEX_BLOCK_SIZE: tl.constexpr = (BLOCK_SIZE + N - 1) // N

    step = num_jobs * INDEX_BLOCK_SIZE
    iters = tl.cdiv(in_n_elements, step)

    # TODO: remove dtype_size once contiguous DMA is ensured
    small_out = inp_numel.to(tl.int64) * dtype_size <= 2**31

    for i in tl.range(iters):
        iter_start = i * step
        iter_end = iter_start + step

        if iter_end <= in_n_elements:
            block_offset = iter_start + pid * INDEX_BLOCK_SIZE
            block_len = INDEX_BLOCK_SIZE
        else:
            rem_n_elements = in_n_elements - iter_start
            base_num = rem_n_elements // num_jobs
            remn_num = rem_n_elements % num_jobs
            extra_one = pid < remn_num

            block_offset = iter_start + (
                (base_num + 1) * pid if extra_one else (base_num * pid + remn_num)
            )
            block_len = base_num + extra_one

        in_offsets = block_offset + tl.arange(0, INDEX_BLOCK_SIZE)
        in_mask = in_offsets < (block_offset + block_len)
        indices = tl.load(in_ptr + in_offsets, in_mask, other=0.0)
        if indices.dtype != tl.int32 and small_out:
            indices_int32 = indices.to(tl.int32)
            ld_st_1(indices_int32, N, weight_ptr, in_mask, in_offsets, out_ptr)
        else:
            ld_st_1(indices, N, weight_ptr, in_mask, in_offsets, out_ptr)


def config_prune(configs, named_args, **kwargs):
    # TODO: bad perf when BLOCK_BATCH is 1
    batch_dim = max(named_args["batch_dim"], 2)
    index_dim = named_args["index_dim"]
    c_dim = named_args["c_dim"]
    dtype_size = named_args["dtype_size"]

    # difficult to include these critical configs while keeping number of configs small
    lb_block_batch_1 = triton.cdiv(batch_dim, TOTAL_CORE_NUM)
    lb_block_batch_2 = max(batch_dim // TOTAL_CORE_NUM, 1)
    lb_block_index_1 = triton.cdiv(index_dim, TOTAL_CORE_NUM)
    lb_block_index_2 = max(index_dim // TOTAL_CORE_NUM, 1)

    max_bs = get_max_block_size(dtype_size)

    block_batches = set([lb_block_batch_1, lb_block_batch_2, batch_dim])
    block_indices = set([lb_block_index_1, lb_block_index_2, index_dim])
    block_cs = set([c_dim, min(max_bs, c_dim)])

    new_configs = []
    for config in configs:
        block_batch = config.kwargs["BLOCK_BATCH"]
        block_index = config.kwargs["BLOCK_INDEX"]
        block_c = config.kwargs["BLOCK_C"]

        # to keep the autotune space small: if c_dim is not very large, don't split c
        block_c_max = 2048 * 5
        block_c = c_dim if c_dim <= block_c_max else block_c

        if block_batch <= batch_dim and block_index <= index_dim and block_c <= c_dim:
            block_batches.add(block_batch)
            block_indices.add(block_index)
            block_cs.add(block_c)

    for block_batch in block_batches:
        for block_index in block_indices:
            for block_c in block_cs:
                if block_batch * block_index * block_c <= max_bs:
                    new_configs.append(
                        triton.Config(
                            {
                                "BLOCK_BATCH": block_batch,
                                "BLOCK_INDEX": block_index,
                                "BLOCK_C": block_c,
                            },
                            num_warps=1,
                            num_stages=1,
                        )
                    )
    return new_configs


@triton.jit
def ld_st_2(
    inp,
    out,
    batch_offsets,
    index_offsets,
    c_offsets,
    inp_strides_0,
    inp_strides_1,
    out_strides_0,
    out_strides_1,
    index_cur,
    input_output_mask,
):
    input_offsets = (batch_offsets * inp_strides_0)[:, None, None] + (
        (index_cur * inp_strides_1)[:, None] + c_offsets[None, :]
    )[None, :, :]

    output_offsets = (batch_offsets * out_strides_0)[:, None, None] + (
        (index_offsets * out_strides_1)[:, None] + c_offsets[None, :]
    )[None, :, :]

    selected = tl.load(inp + input_offsets, mask=input_output_mask, other=0.0)
    tl.store(out + output_offsets, selected, mask=input_output_mask)


@libentry()
@libtuner(
    configs=runtime.get_tuned_config("index_select"),
    key=["batch_dim", "index_dim", "c_dim"],
    prune_configs_by={"early_config_prune": config_prune},
)
@triton.jit
def multi_batch_index_select_kernel(
    inp,
    index,
    out,
    batch_dim,
    select_dim,
    c_dim,
    index_dim,
    dtype_size,
    inp_numel,
    BLOCK_BATCH: tl.constexpr,
    BLOCK_INDEX: tl.constexpr,
    BLOCK_C: tl.constexpr,
):
    pid_x = tl.program_id(axis=0)
    num_programs = tl.num_programs(axis=0)

    block_id_start = pid_x
    block_id_step = num_programs

    block_batch: tl.constexpr = BLOCK_BATCH
    block_index: tl.constexpr = BLOCK_INDEX
    block_c: tl.constexpr = BLOCK_C

    block_num_batch = tl.cdiv(batch_dim, block_batch)
    block_num_index = tl.cdiv(index_dim, block_index)
    block_num_c = tl.cdiv(c_dim, block_c)

    block_num_total = block_num_batch * block_num_index * block_num_c

    inp_strides_0, inp_strides_1 = [select_dim * c_dim, c_dim]
    out_strides_0, out_strides_1 = [index_dim * c_dim, c_dim]
    block_strides_0, block_strides_1 = [block_num_index * block_num_c, block_num_c]

    # TODO: remove dtype_size once contiguous DMA is ensured
    small_out = inp_numel.to(tl.int64) * dtype_size <= 2**31

    for block_id in tl.range(block_id_start, block_num_total, block_id_step):
        block_id_batch = block_id // block_strides_0
        block_id_index = (block_id // block_strides_1) % block_num_index
        block_id_c = block_id % block_num_c

        # arange requires constexpr
        batch_offsets = block_id_batch * block_batch + tl.arange(0, block_batch)
        batch_mask = batch_offsets < batch_dim

        index_offsets = block_id_index * block_index + tl.arange(0, block_index)
        index_mask = index_offsets < index_dim

        c_offsets = block_id_c * block_c + tl.arange(0, block_c)
        c_mask = c_offsets < c_dim

        input_output_mask = (
            batch_mask[:, None, None]
            and (index_mask[:, None] and c_mask[None, :])[None, :, :]
        )

        index_cur = tl.load(index + index_offsets, mask=index_mask, other=0)
        # TODO: remove dtype_size once contiguous DMA is ensured
        if index.dtype != tl.int32 and small_out:
            index_cur_int32 = index_cur.to(tl.int32)
            ld_st_2(
                inp,
                out,
                batch_offsets,
                index_offsets,
                c_offsets,
                inp_strides_0,
                inp_strides_1,
                out_strides_0,
                out_strides_1,
                index_cur_int32,
                input_output_mask,
            )
        else:
            ld_st_2(
                inp,
                out,
                batch_offsets,
                index_offsets,
                c_offsets,
                inp_strides_0,
                inp_strides_1,
                out_strides_0,
                out_strides_1,
                index_cur,
                input_output_mask,
            )


def index_select(inp, dim, index):
    logger.debug("GEMS_CAMBRICON INDEX SELECT")
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    assert index.ndim <= 1, "Index should have dimension 1 or 0"
    # TODO: index is on device, should it be a kernel (like cnnl __assert_fail__) to check this?
    assert ((i >= 0 and i < inp.size(dim)) for i in index), "Index out of range"

    # TODO: make sure input is contiguous

    if index.ndim == 0:
        index = index.unsqueeze(0)
    dim = dim % inp.ndim
    inp_shape = list(inp.shape)
    index_dim = index.numel()

    # input  [batch_dim, select_dim, c_dim]
    # output [batch_dim, index_dim, c_dim]
    inp = inp.contiguous()
    index = index.contiguous()
    inp_numel = inp.numel()
    batch_dim = math.prod(inp_shape[:dim])
    select_dim = inp_shape[dim]
    c_dim = math.prod(inp_shape[(dim + 1) :])

    out_shape = inp_shape
    out_shape[dim] = index_dim
    out = torch.empty(out_shape, dtype=inp.dtype, device=inp.device)

    if torch.is_floating_point(inp):
        dtype_size = torch.finfo(inp.dtype).bits // 8
    else:
        dtype_size = torch.iinfo(inp.dtype).bits // 8

    if batch_dim == 1 and c_dim <= get_max_block_size(dtype_size):
        # ram: (input, output), half, extra
        # 2D, not split c_dim
        def grid_fn(meta):
            index_block_size_grid = max(meta["BLOCK_SIZE"] // c_dim, 1)
            index_block_num = triton.cdiv(index_dim, index_block_size_grid)
            return (min(index_block_num, TOTAL_CORE_NUM),)

        one_batch_index_select_kernel[grid_fn](
            out, index, index_dim, inp, c_dim, dtype_size, inp_numel
        )
    else:
        grid = lambda meta: (
            min(
                triton.cdiv(batch_dim, meta["BLOCK_BATCH"])
                * triton.cdiv(index_dim, meta["BLOCK_INDEX"])
                * triton.cdiv(c_dim, meta["BLOCK_C"]),
                TOTAL_CORE_NUM,
            ),
        )
        multi_batch_index_select_kernel[grid](
            inp,
            index,
            out,
            batch_dim,
            select_dim,
            c_dim,
            index_dim,
            dtype_size,
            inp_numel,
        )
    return out



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch
from einops import rearrange

from fla.modules.l2norm import l2norm_bwd, l2norm_fwd
from fla.ops.common.chunk_scaled_dot_kkt import chunk_scaled_dot_kkt_fwd
from fla.ops.delta_rule.chunk import chunk_delta_rule_bwd
from fla.ops.delta_rule.wy_fast import recompute_w_u_fwd as dn_recompute_w_u_fwd
from fla.ops.gated_delta_product.chunk_deltaproduct_h import chunk_gated_delta_product_fwd_h
from fla.ops.gated_delta_product.chunk_deltaproduct_o import chunk_gated_delta_product_fwd_o
from fla.ops.gated_delta_rule.chunk import chunk_gated_delta_rule_bwd
from fla.ops.gated_delta_rule.wy_fast import recompute_w_u_fwd as gdn_recompute_w_u_fwd
from fla.ops.utils import chunk_local_cumsum, solve_tril
from fla.utils import autocast_custom_bwd, autocast_custom_fwd, input_guard


def chunk_gated_delta_product_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    beta: torch.Tensor,
    scale: float,
    cu_seqlens: torch.LongTensor | None = None,
    initial_state: torch.Tensor | None = None,
    output_final_state: bool = False,
    num_householder: int = 1,
):
    cu_seqlens_dp = cu_seqlens * num_householder if cu_seqlens is not None else None
    if g is not None:
        g_interleaved = g.new_zeros(g.shape[0], g.shape[1], num_householder, g.shape[2], dtype=torch.float32)
        g_interleaved[:, :, 0] = g
        g_interleaved = rearrange(g_interleaved, 'b l n h -> b (l n) h').contiguous()
        g = chunk_local_cumsum(g, chunk_size=64, cu_seqlens=cu_seqlens, output_dtype=torch.float32)
        g_interleaved = chunk_local_cumsum(g_interleaved, chunk_size=64, cu_seqlens=cu_seqlens_dp, output_dtype=torch.float32)
    else:
        g_interleaved = None
        g = None
    # obtain WY representation. u is actually the new v.
    A = chunk_scaled_dot_kkt_fwd(
        k=k,
        g=g_interleaved,
        beta=beta,
        cu_seqlens=cu_seqlens_dp,
        output_dtype=torch.float32,
    )
    A = solve_tril(
        A=A,
        cu_seqlens=cu_seqlens_dp,
        output_dtype=k.dtype,
    )
    if g is not None:
        w, u = gdn_recompute_w_u_fwd(
            k=k,
            v=v,
            beta=beta,
            A=A,
            g=g_interleaved,
            cu_seqlens=cu_seqlens_dp,
        )
    else:
        w, u = dn_recompute_w_u_fwd(
            k=k,
            v=v,
            beta=beta,
            A=A,
            cu_seqlens=cu_seqlens_dp,
        )
    h, v_new, final_state = chunk_gated_delta_product_fwd_h(
        k=k,
        w=w,
        u=u,
        g=g_interleaved,
        initial_state=initial_state,
        output_final_state=output_final_state,
        cu_seqlens=cu_seqlens_dp,
        num_householder=num_householder,
    )
    o = chunk_gated_delta_product_fwd_o(
        q=q,
        k=k,
        v=v_new,
        h=h,
        g=g,
        scale=scale,
        cu_seqlens=cu_seqlens,
        num_householder=num_householder,
    )
    return g, g_interleaved, o, A, final_state


class ChunkGatedDeltaProductFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    @autocast_custom_fwd
    def forward(
        ctx,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        g: torch.Tensor,
        beta: torch.Tensor,
        scale: float,
        num_householder: int,
        initial_state: torch.Tensor,
        output_final_state: bool,
        use_qk_l2norm_in_kernel: bool = False,
        cu_seqlens: torch.LongTensor | None = None,
    ):
        if use_qk_l2norm_in_kernel:
            q, q_rstd = l2norm_fwd(q)
            k, k_rstd = l2norm_fwd(k)
        else:
            q_rstd, k_rstd = None, None

        g, g_interleaved, o, A, final_state = chunk_gated_delta_product_fwd(
            q=q,
            k=k,
            v=v,
            g=g,
            beta=beta,
            scale=scale,
            initial_state=initial_state,
            output_final_state=output_final_state,
            cu_seqlens=cu_seqlens,
            num_householder=num_householder,
        )
        ctx.save_for_backward(q, q_rstd, k, k_rstd, v, g_interleaved, beta, A, initial_state, cu_seqlens)
        ctx.scale = scale
        ctx.use_qk_l2norm_in_kernel = use_qk_l2norm_in_kernel
        ctx.num_householder = num_householder
        return o.to(q.dtype), final_state

    @staticmethod
    @input_guard
    @autocast_custom_bwd
    def backward(
        ctx,
        do: torch.Tensor,
        dht: torch.Tensor,
    ):
        q, q_rstd, k, k_rstd, v, g, beta, A, initial_state, cu_seqlens = ctx.saved_tensors
        q_new = q.new_zeros(q.shape[0], q.shape[1], ctx.num_householder, q.shape[2], q.shape[3])
        q_new[:, :, -1] = q
        do_new = do.new_zeros(do.shape[0], do.shape[1], ctx.num_householder, do.shape[2], do.shape[3])
        do_new[:, :, -1] = do
        q_org, q = q, rearrange(q_new, 'b t n h d -> b (t n) h d')
        do = rearrange(do_new, 'b t n h d -> b (t n) h d')
        # call the gated deltanet kernel for now.
        # TODO: optimize the backward pass like the forward pass.
        if g is not None:
            dq, dk, dv, db, dg, dh0 = chunk_gated_delta_rule_bwd(
                q=q,
                k=k,
                v=v,
                g=g,
                beta=beta,
                A=A,
                scale=ctx.scale,
                initial_state=initial_state,
                do=do,
                dht=dht,
                cu_seqlens=cu_seqlens * ctx.num_householder if cu_seqlens is not None else None,
            )
            dg = rearrange(dg, 'b (l n) h  -> b l n h ', n=ctx.num_householder)[:, :, 0].contiguous().to(g)
        else:
            dq, dk, dv, db, dh0 = chunk_delta_rule_bwd(
                q=q,
                k=k,
                v=v,
                beta=beta,
                A=A,
                scale=ctx.scale,
                initial_state=initial_state,
                do=do,
                dht=dht,
                cu_seqlens=cu_seqlens * ctx.num_householder if cu_seqlens is not None else None,
            )
            dg = None
        dq = rearrange(dq, 'b (l n) h d -> b l n h d', n=ctx.num_householder)[:, :, -1].contiguous()
        if ctx.use_qk_l2norm_in_kernel:
            dq = l2norm_bwd(q_org, q_rstd, dq)
            dk = l2norm_bwd(k, k_rstd, dk)
        return dq.to(q), dk.to(k), dv.to(v), dg, db.to(beta), None, None, dh0, None, None, None


@torch.compiler.disable
def chunk_gated_delta_product(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    beta: torch.Tensor,
    num_householder: int,
    scale: float = None,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    use_qk_l2norm_in_kernel: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
):
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, H, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        g (torch.Tensor):
            (forget) gating tensor (in log space!) of shape `[B, T, H]`.
        beta (torch.Tensor):
            betas of shape `[B, T, H]`.
        num_householder (int):
            Number of householder transformations to apply. Default: `1`.
        scale (Optional[float]):
            Scale factor for the RetNet attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `[N, H, K, V]` for `N` input sequences.
            For equal-length input sequences, `N` equals the batch size `B`.
            Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `[N, H, K, V]`. Default: `False`.
        use_qk_l2norm_in_kernel (Optional[bool]):
            Whether to use qk l2norm within the kernel for saving GPU memory.
            Default: `False`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, H, V]`.
        final_state (torch.Tensor):
            Final state of shape `[N, H, K, V]` if `output_final_state=True` else `None`.

    Examples::
        >>> import torch
        >>> import torch.nn.functional as F
        >>> from einops import rearrange
        >>> from fla.ops.gated_delta_rule import chunk_gated_delta_product
        # inputs with equal lengths
        >>> B, T, H, K, V = 4, 2048, 4, 512, 512
        >>> q = torch.randn(B, T, H, K, dtype=torch.bfloat16, device='cuda')
        >>> k = F.normalize(torch.randn(B, T, H, K, dtype=torch.bfloat16, device='cuda'), p=2, dim=-1)
        >>> v = torch.randn(B, T, H, V, dtype=torch.bfloat16, device='cuda')
        >>> beta = torch.rand(B, T, H, dtype=torch.bfloat16, device='cuda').sigmoid()
        >>> g = F.logsigmoid(torch.rand(B, T, H, dtype=torch.bfloat16, device='cuda'))
        >>> h0 = torch.randn(B, H, K, V, dtype=torch.bfloat16, device='cuda')
        >>> o, ht = chunk_gated_delta_product(
            q, k, v, g, beta,
            initial_state=h0,
            output_final_state=True
        )
        # for variable-length inputs, the batch size `B` is expected to be 1 and `cu_seqlens` is required
        >>> q, k, v, beta, g = map(lambda x: rearrange(x, 'b t ... -> 1 (b t) ...'), (q, k, v, beta, g))
        # for a batch with 4 sequences, `cu_seqlens` with 5 start/end positions are expected
        >>> cu_seqlens = q.new_tensor([0, 2048, 4096, 6144, 8192], dtype=torch.long)
        >>> o, ht = chunk_gated_delta_product(
            q, k, v, g, beta,
            initial_state=h0,
            output_final_state=True,
            cu_seqlens=cu_seqlens
        )
    """
    assert q.dtype != torch.float32, "ChunkGatedDeltaProductFunction does not support float32. Please use bfloat16."
    B, T, H, K, V = *q.shape, v.shape[-1]
    assert k.shape == (B, T*num_householder, H, K)
    assert v.shape == (B, T*num_householder, H, V)
    assert beta.shape == (B, T*num_householder, H)
    if g is not None:
        assert g.shape == (B, T, H)

    if cu_seqlens is not None:
        if q.shape[0] != 1:
            raise ValueError(
                f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
                f"Please flatten variable-length inputs before processing.",
            )
        if initial_state is not None and initial_state.shape[0] != len(cu_seqlens) - 1:
            raise ValueError(
                f"The number of initial states is expected to be equal to the number of input sequences, "
                f"i.e., {len(cu_seqlens) - 1} rather than {initial_state.shape[0]}.",
            )
    if scale is None:
        scale = k.shape[-1] ** -0.5
    o, final_state = ChunkGatedDeltaProductFunction.apply(
        q,
        k,
        v,
        g,
        beta,
        scale,
        num_householder,
        initial_state,
        output_final_state,
        use_qk_l2norm_in_kernel,
        cu_seqlens,
    )
    return o, final_state



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle
from flag_gems.utils.limits import get_dtype_min

logger = logging.getLogger(__name__)


@libentry()
@triton.jit
def argmax_kernel_1(
    inp,
    mid_value,
    mid_index,
    M,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    min_value = get_dtype_min(inp.type.element_ty)
    inp_val = tl.load(inp_ptrs, mask=mask, other=min_value)
    max_val, max_index = tl.max(inp_val, axis=0, return_indices=True)
    max_index = max_index + pid * BLOCK_SIZE
    mid_value_ptr = mid_value + pid
    max_index_ptr = mid_index + pid
    tl.store(mid_value_ptr, max_val)
    tl.store(max_index_ptr, max_index)


@libentry()
@triton.jit
def argmax_kernel_2(mid_value, mid_index, out, mid_size, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid_value + offset
    mask = offset < mid_size
    min_value = get_dtype_min(mid_value.type.element_ty)
    mid_val = tl.load(mid_ptrs, mask=mask, other=min_value)
    index_val = tl.argmax(mid_val, axis=0)
    mid_index_ptrs = mid_index + index_val
    out_val = tl.load(mid_index_ptrs)
    tl.store(out, out_val)


@libentry()
@triton.heuristics(runtime.get_heuristic_config("argmax_non_inner"))
@triton.jit
def argmax_kernel_non_inner(
    inp,
    out_index,
    M,
    N,
    K,
    TILE_K: tl.constexpr,
    TILE_N: tl.constexpr,
    ONE_TILE_PER_CTA: tl.constexpr,
):
    pid_m = tle.program_id(0)
    pid_k = tle.program_id(1)
    k_offset = pid_k * TILE_K + tl.arange(0, TILE_K)

    if tl.constexpr(inp.dtype.element_ty == tl.float16) or tl.constexpr(
        inp.dtype.element_ty == tl.bfloat16
    ):
        cdtype = tl.float32
    else:
        cdtype = inp.dtype.element_ty

    min_value = get_dtype_min(cdtype)

    if ONE_TILE_PER_CTA:
        n_offset = tl.arange(0, TILE_N)
        offset = pid_m * N * K + n_offset[:, None] * K + k_offset
        mask = k_offset < K and n_offset[:, None] < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
        local_max, local_argmax = tl.max(
            inp_vals, 0, return_indices=True, return_indices_tie_break_left=True
        )
        offset_index = pid_m * K + k_offset
        out_index_ptrs = out_index + offset_index
        mask1 = k_offset < K
        tl.store(out_index_ptrs, local_argmax, mask=mask1)
    else:
        max_values = tl.full([TILE_K], dtype=cdtype, value=min_value)
        argmax_values = tl.full([TILE_K], dtype=tl.int64, value=0)

        for start_n in range(0, N, TILE_N):
            n_offset = start_n + tl.arange(0, TILE_N)
            offset = pid_m * N * K + n_offset[:, None] * K + k_offset
            mask = k_offset < K and n_offset[:, None] < N
            inp_ptrs = inp + offset
            inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
            local_max, local_argmax = tl.max(
                inp_vals, 0, return_indices=True, return_indices_tie_break_left=True
            )
            update = local_max > max_values
            max_values = tl.where(update, local_max, max_values)
            argmax_values = tl.where(update, start_n + local_argmax, argmax_values)
        offset_index = pid_m * K + k_offset
        out_index_ptrs = out_index + offset_index
        mask1 = k_offset < K
        tl.store(out_index_ptrs, argmax_values, mask=mask1)


@libentry()
@triton.heuristics(runtime.get_heuristic_config("argmax_inner"))
@triton.jit
def argmax_kernel_inner(
    inp,
    out_index,
    M,
    N,
    TILE_N: tl.constexpr,
    ONE_TILE_PER_CTA: tl.constexpr,
):
    pid_m = tle.program_id(0)

    dtype = inp.type.element_ty
    min_value = get_dtype_min(dtype)

    if ONE_TILE_PER_CTA:
        n_offset = tl.arange(0, TILE_N)
        offset = pid_m * N + n_offset
        mask = n_offset < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
        local_max, local_argmax = tl.max(
            inp_vals, 0, return_indices=True, return_indices_tie_break_left=True
        )
        out_index_ptrs = out_index + pid_m
        tl.store(out_index_ptrs, local_argmax)
    else:
        max_values = min_value
        argmax_values = 0

        loop_time = N // TILE_N
        remainder = N % TILE_N
        for start_n in range(0, loop_time):
            n_offset = start_n * TILE_N + tl.arange(0, TILE_N)
            offset = pid_m * N + n_offset
            inp_ptrs = inp + offset
            inp_vals = tl.load(inp_ptrs)
            local_max, local_argmax = tl.max(
                inp_vals, 0, return_indices=True, return_indices_tie_break_left=True
            )
            update = local_max > max_values
            max_values = tl.where(update, local_max, max_values)
            argmax_values = tl.where(
                update, start_n * TILE_N + local_argmax, argmax_values
            )

        if remainder:
            n_offset = loop_time * TILE_N + tl.arange(0, TILE_N)
            offset = pid_m * N + n_offset
            mask = n_offset < N
            inp_ptrs = inp + offset
            inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
            local_max, local_argmax = tl.max(
                inp_vals, 0, return_indices=True, return_indices_tie_break_left=True
            )
            update = local_max > max_values
            max_values = tl.where(update, local_max, max_values)
            argmax_values = tl.where(
                update, loop_time * TILE_N + local_argmax, argmax_values
            )

        out_index_ptrs = out_index + pid_m
        tl.store(out_index_ptrs, argmax_values)


def argmax(inp, dim=None, keepdim=False, *, dtype=None):
    logger.debug("GEMS ARGMAX")
    if dim is None:
        M = inp.numel()
        if dtype is None:
            dtype = inp.dtype
        block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
        mid_size = triton.cdiv(M, block_size)
        block_mid = triton.next_power_of_2(mid_size)

        mid_value = torch.empty((mid_size,), dtype=dtype, device=inp.device)
        mid_index = torch.empty((mid_size,), dtype=torch.int64, device=inp.device)
        if keepdim:
            shape = list(inp.shape)
            for i in range(0, inp.dim()):
                shape[i] = 1
            out = torch.empty(shape, dtype=torch.int64, device=inp.device)
        else:
            out = torch.empty([], dtype=torch.int64, device=inp.device)

        with torch_device_fn.device(inp.device):
            argmax_kernel_1[(mid_size, 1, 1)](
                inp,
                mid_value,
                mid_index,
                M,
                block_size,
            )
            argmax_kernel_2[(1, 1, 1)](mid_value, mid_index, out, mid_size, block_mid)
        return out
    else:
        assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
        shape = inp.shape
        dim = dim % inp.ndim
        if inp.numel() == 0:
            out_shape = list(shape)
            if keepdim:
                out_shape[dim] = 1
            else:
                del out_shape[dim]
            return torch.zeros(out_shape, dtype=torch.int64, device=inp.device)
        N = shape[dim]
        M = math.prod(shape[:dim])
        K = inp.numel() // M // N

        inp = inp.contiguous()

        shape_list = list(shape)
        shape_list[dim] = 1
        out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)
        if not keepdim:
            out_index = torch.squeeze(out_index, dim)

        with torch_device_fn.device(inp.device):
            if K > 1:
                grid = lambda meta: (
                    M,
                    triton.cdiv(K, meta["TILE_K"]),
                )
                argmax_kernel_non_inner[grid](
                    inp,
                    out_index,
                    M,
                    N,
                    K,
                )
            else:
                grid = lambda meta: (M, 1, 1)
                argmax_kernel_inner[grid](
                    inp,
                    out_index,
                    M,
                    N,
                )
        return out_index



"""
Reshape - Reshape tensor to new shape
Type: Metadata Op (ZERO compute if contiguous, copy otherwise)

NeuroBrix - NVIDIA Common (All architectures)
"""
import torch
from typing import List, Union

from neurobrix.kernels.registry import register_kernel


@register_kernel(family="audio", vendor="nvidia", tier="common", op_name="reshape")
def reshape_kernel(x: torch.Tensor, shape: Union[List[int], torch.Size]) -> torch.Tensor:
    """
    Reshape - Reshape tensor to new shape.
    
    Uses torch.reshape():
    - If tensor is contiguous and new shape is compatible: returns view (ZERO compute)
    - Otherwise: returns contiguous copy
    
    Supports -1 for automatic dimension inference.
    """
    return x.reshape(*shape)


import logging

import torch
import triton
from _kunlunxin.ops.copy import copy_slice

from flag_gems.utils.shape_utils import MemOverlap, has_internal_overlapping

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


def slice_scatter(inp, src, dim=0, start=None, end=None, step=1):
    logger.debug("GEMS SLICE_SCATTER")
    assert src.device == inp.device, "inp and src reside on different devices."
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    assert step > 0, "slice step must be positive"
    dim = dim % inp.ndim

    start = start or 0
    end = end or inp.size(dim)
    if end < 0:
        end = end % inp.size(dim)

    valid_shape = list(inp.shape)
    valid_shape[dim] = triton.cdiv(end - start, step)
    assert (
        list(src.shape) == valid_shape
    ), "Expected src to have a size equal to the slice of self"

    if has_internal_overlapping(inp) == MemOverlap.Yes:
        out = torch.empty(inp.size(), dtype=inp.dtype, device=inp.device)
    else:
        out = torch.empty_strided(
            inp.size(), inp.stride(), dtype=inp.dtype, device=inp.device
        )

    ndim = inp.ndim
    copy_slice(inp, out0=out)

    indices = [slice(None)] * ndim
    indices[dim] = slice(start, end, step)
    out_ = out[indices]
    copy_slice(src, out0=out_)

    return out



import logging

import torch
import triton

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(
    is_tensor=[True, True, False, False], promotion_methods=[(0, 1, "DEFAULT")]
)
@triton.jit
def add_func(x, y, alpha, inplace):
    return x + y * alpha


@pointwise_dynamic(
    is_tensor=[True, False, False, False], promotion_methods=[(0, 1, "DEFAULT")]
)
@triton.jit
def add_func_tensor_scalar(x, y, alpha, inplace):
    return x + y * alpha


@pointwise_dynamic(
    is_tensor=[False, True, False, False], promotion_methods=[(0, 1, "DEFAULT")]
)
@triton.jit
def add_func_scalar_tensor(x, y, alpha, inplace):
    return x + y * alpha


def add(A, B, *, alpha=1):
    logger.debug("GEMS_CAMBRICON ADD")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return add_func(A, B, alpha, False)
    elif isinstance(A, torch.Tensor):
        return add_func_tensor_scalar(A, B, alpha, False)
    elif isinstance(B, torch.Tensor):
        return add_func_scalar_tensor(A, B, alpha, False)
    else:
        return torch.tensor(A + B * alpha)


def add_(A, B, *, alpha=1):
    logger.debug("GEMS_CAMBRICON ADD_")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return add_func(A, B, alpha, True, out0=A)
    elif isinstance(A, torch.Tensor):
        return add_func_tensor_scalar(A, B, alpha, True, out0=A)
    # elif isinstance(B, torch.Tensor):
    #     return add_func_scalar_tensor(A, B, alpha, out0=A)
    else:
        raise ValueError("Unreachable.")



"""
Expand - Broadcast tensor to new shape
ATen: aten::expand
"""
import torch
from typing import List

def expand_kernel(x: torch.Tensor, size: list = None) -> torch.Tensor:
    """Expand tensor to new size."""
    if size is None:
        return x
    return x.expand(*size)



import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import broadcastable_to, libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(__name__)


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("mv"),
    key=["M", "N"],
)
@triton.jit(do_not_specialize=["alpha", "beta"])
def addmv_kernel(
    A,
    B,
    Inp,
    Out,
    N,
    M,
    alpha,
    beta,
    stride_an,
    stride_am,
    stride_bm,
    stride_in,
    stride_outn,
    BLOCK_N: tl.constexpr,
    BLOCK_M: tl.constexpr,
):
    pid = tle.program_id(0)
    offset_n = pid * BLOCK_N + tl.arange(0, BLOCK_N)[:, None]
    offset_m = tl.arange(0, BLOCK_M)[None, :]
    n_mask = offset_n < N
    A_ptrs = A + offset_n * stride_an + offset_m * stride_am
    B_ptrs = B + offset_m * stride_bm
    acc = tl.zeros((BLOCK_N, BLOCK_M), dtype=tl.float32)
    for m in range(0, M, BLOCK_M):
        m_mask = m + offset_m < M
        a = tl.load(A_ptrs, mask=n_mask & m_mask, other=0.0).to(tl.float32)
        b = tl.load(B_ptrs, mask=m_mask, other=0.0).to(tl.float32)
        acc += a * b
        A_ptrs += BLOCK_M * stride_am
        B_ptrs += BLOCK_M * stride_bm

    acc = tl.sum(acc, axis=1)[:, None]
    Inp_ptrs = Inp + offset_n * stride_in
    inp = tl.load(Inp_ptrs, mask=n_mask, other=0.0).to(tl.float32)
    Out_ptrs = Out + offset_n * stride_outn
    out_block = acc * alpha + inp * beta
    tl.store(Out_ptrs, out_block, mask=n_mask)


def addmv(self, mat, vec, *, beta=1, alpha=1):
    logger.debug("GEMS ADDMV")
    assert mat.shape[1] == vec.shape[0], "incompatible dimensions"
    assert broadcastable_to(self.shape, (mat.shape[0],)), "Incompatible self shape"
    N, M = mat.shape
    out = torch.empty((N,), device=mat.device, dtype=mat.dtype)
    self = self.broadcast_to(out.shape)
    grid = lambda META: (triton.cdiv(N, META["BLOCK_N"]),)
    with torch_device_fn.device(mat.device):
        addmv_kernel[grid](
            mat,
            vec,
            self,
            out,
            N,
            M,
            alpha,
            beta,
            mat.stride(0),
            mat.stride(1),
            vec.stride(0),
            self.stride(0),
            out.stride(0),
        )
    return out


def addmv_out(self, mat, vec, *, beta=1, alpha=1, out=None):
    logger.debug("GEMS ADDMV OUT")
    assert mat.shape[1] == vec.shape[0], "incompatible dimensions"
    assert broadcastable_to(self.shape, (mat.shape[0],)), "Incompatible self shape"
    N, M = mat.shape
    if out is None:
        out = torch.empty((N,), device=mat.device, dtype=mat.dtype)
    else:
        assert out.shape == (N,), "Incompatible output shape"

    self = self.broadcast_to(out.shape)
    grid = lambda META: (triton.cdiv(N, META["BLOCK_N"]),)
    with torch_device_fn.device(mat.device):
        addmv_kernel[grid](
            mat,
            vec,
            self,
            out,
            N,
            M,
            alpha,
            beta,
            mat.stride(0),
            mat.stride(1),
            vec.stride(0),
            self.stride(0),
            out.stride(0),
        )
    return out



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
#
# modified from https://github.com/mdy666/mdy_triton/blob/e0a856347bd988e05e0152332bba35f1d33c5b1f/others/grpo/grpo_loss.ipynb
# XHS ID: blueeeee

# https://github.com/huggingface/trl/blob/main/trl/trainer/grpo_trainer.py
"""
# Get the per-token log probabilities for the completions for the model and the reference model
    def _get_per_token_logps(self, model, input_ids, attention_mask, logits_to_keep):
        # We add 1 to `logits_to_keep` because the last logits of the sequence is later excluded
        logits = model(input_ids=input_ids, attention_mask=attention_mask, logits_to_keep=logits_to_keep + 1).logits
        logits = logits[:, :-1, :]  # (B, L-1, V), exclude the last logit: it corresponds to the next token pred

        input_ids = input_ids[:, -logits_to_keep:]
        # For transformers<=4.48, logits_to_keep argument isn't supported, so here we drop logits ourselves.
        # See https://github.com/huggingface/trl/issues/2770
        logits = logits[:, -logits_to_keep:]
        return selective_log_softmax(logits, input_ids)  #  compute logprobs for the input tokens

    def compute_loss(self, model, inputs, return_outputs=False, num_items_in_batch=None):
        if return_outputs:
            raise ValueError("The GRPOTrainer does not support returning outputs")
        # Compute the per-token log probabilities for the model

        prompt_ids, prompt_mask = inputs["prompt_ids"], inputs["prompt_mask"]
        completion_ids, completion_mask = inputs["completion_ids"], inputs["completion_mask"]
        input_ids = torch.cat([prompt_ids, completion_ids], dim=1)
        attention_mask = torch.cat([prompt_mask, completion_mask], dim=1)
        logits_to_keep = completion_ids.size(1)  # we only need to compute the logits for the completion tokens

        per_token_logps = self._get_per_token_logps(model, input_ids, attention_mask, logits_to_keep)

        # Compute the KL divergence between the model and the reference model
        ref_per_token_logps = inputs["ref_per_token_logps"]
        per_token_kl = torch.exp(ref_per_token_logps - per_token_logps) - (ref_per_token_logps - per_token_logps) - 1

        # x - x.detach() allows for preserving gradients from x
        advantages = inputs["advantages"]
        per_token_loss = torch.exp(per_token_logps - per_token_logps.detach()) * advantages.unsqueeze(1)
        per_token_loss = -(per_token_loss - self.beta * per_token_kl)
        loss = ((per_token_loss * completion_mask).sum(dim=1) / completion_mask.sum(dim=1)).mean()

        # Log the metrics
        completion_length = self.accelerator.gather_for_metrics(completion_mask.sum(1)).float().mean().item()
        self._metrics["completion_length"].append(completion_length)

        mean_kl = ((per_token_kl * completion_mask).sum(dim=1) / completion_mask.sum(dim=1)).mean()
        self._metrics["kl"].append(self.accelerator.gather_for_metrics(mean_kl).mean().item())

        return loss
"""


import torch
import triton
import triton.language as tl

from fla.ops.utils.op import exp, log
from fla.utils import IS_AMD, autotune_cache_kwargs, input_guard

NUM_WARPS_AUTOTUNE = [4, 8, 16] if IS_AMD else [4, 8, 16, 32]


@triton.autotune(
    configs=[
        triton.Config({'BLOCK_SIZE': BLOCK_SIZE},  num_warps=NUM_WARPS, num_stages=NUM_STAGES)
        for BLOCK_SIZE in [1024, 2048, 4096, 8192]
        for NUM_WARPS in NUM_WARPS_AUTOTUNE
        for NUM_STAGES in [1, 2, 4]
    ],
    key=['B', 'N'],
    **autotune_cache_kwargs,
)
@triton.jit
def grpo_fwd_kernel(
    logits_ptr,
    ref_logp_ptr,
    input_ids_ptr,
    advantages_ptr,
    completion_mask_ptr,
    loss_ptr,
    lse_ptr,
    beta,
    save_kl: tl.constexpr,
    B,
    M,
    N,
    L,
    start_idx,
    BLOCK_SIZE: tl.constexpr,
):
    row_idx = tl.program_id(0)

    off_b = row_idx // L
    N = tl.cast(N, tl.int64)

    loss_ptr += row_idx

    completion_mask_ptr += row_idx
    not_skip = tl.load(completion_mask_ptr).to(tl.int1)
    if not_skip == 1:
        ref_logp_ptr += row_idx
        lse_ptr += row_idx
        advantages_ptr += off_b
        logits_ptr += N * (row_idx + off_b)
        input_ids_ptr += row_idx + (off_b+1) * start_idx
        base_cols = tl.arange(0, BLOCK_SIZE)

        m_i = -float("inf")
        l_i = 0.0
        for start_n in tl.range(0, N, BLOCK_SIZE):
            cols = start_n + base_cols
            mask = cols < N
            logits = tl.load(logits_ptr+cols, mask=mask, other=-float('inf')).to(tl.float32)
            m_ij = tl.max(logits)
            new_m_i = tl.maximum(m_i, m_ij)
            l_i = l_i * exp(m_i - new_m_i) + tl.sum(exp(logits - new_m_i))
            m_i = new_m_i
        lse = log(l_i) + m_i

        idx = tl.load(input_ids_ptr)
        x = tl.load(logits_ptr+idx).to(tl.float32)
        advantage = tl.load(advantages_ptr).to(tl.float32)
        ref_logp = tl.load(ref_logp_ptr)
        logp = x - lse
        diff = ref_logp - logp
        kl = exp(diff) - diff - 1
        loss = kl * beta - advantage

        tl.store(loss_ptr, loss.to(loss_ptr.dtype.element_ty))
        tl.store(lse_ptr, lse.to(lse_ptr.dtype.element_ty))
        if save_kl:
            tl.store(loss_ptr+M, kl.to(loss_ptr.dtype.element_ty))
    else:
        # store 0
        tl.store(loss_ptr, 0.0)
        if save_kl:
            tl.store(loss_ptr+M, 0.0)


@triton.autotune(
    configs=[
        triton.Config({},  num_warps=NUM_WARPS, num_stages=NUM_STAGES)
        for NUM_WARPS in [32]
        for NUM_STAGES in [4]
    ],
    key=['B', 'N'],
    **autotune_cache_kwargs,
)
@triton.jit
def grpo_bwd_kernel(
    dloss_ptr,
    dlogits_ptr,
    logits_ptr,
    ref_logp_ptr,
    input_ids_ptr,
    advantages_ptr,
    completion_mask_ptr,
    lse_ptr,
    beta,
    B,
    N,
    L,
    start_idx,
    BLOCK_SIZE: tl.constexpr,
):

    row_idx = tl.program_id(0)  # B*L
    off_b = row_idx // L

    N = tl.cast(N, tl.int64)

    dlogits_ptr += N * (row_idx + off_b)
    base_cols = tl.arange(0, BLOCK_SIZE)
    completion_mask_ptr += row_idx
    not_skip = tl.load(completion_mask_ptr).to(tl.int1)

    if not_skip == 1:
        lse_ptr += row_idx
        dloss_ptr += row_idx
        advantages_ptr += off_b
        ref_logp_ptr += row_idx
        logits_ptr += N * (row_idx + off_b)
        input_ids_ptr += row_idx + (off_b+1) * start_idx
        dloss = tl.load(dloss_ptr).to(tl.float32)
        lse = tl.load(lse_ptr).to(tl.float32)
        idx = tl.load(input_ids_ptr)
        x = tl.load(logits_ptr+idx).to(tl.float32)
        advantage = tl.load(advantages_ptr).to(tl.float32)
        ref_logp = tl.load(ref_logp_ptr)
        # Need for in-place grad.
        tl.debug_barrier()
        logp = x - lse

        dlogp = (beta * (-1.0 * exp(ref_logp - logp) + 1)
                 - advantage) * dloss

        for start_n in tl.range(0, N, BLOCK_SIZE):
            cols = start_n + base_cols
            mask = cols < N
            logits = tl.load(logits_ptr+cols, mask=mask, other=-float('inf')).to(tl.float32)
            probs = exp(logits - lse)
            dlogits = tl.where(cols == idx, 1-probs, -probs) * dlogp

            tl.store(dlogits_ptr+cols, dlogits.to(dlogits_ptr.dtype.element_ty), mask=mask)
    else:
        dlogits = tl.zeros((BLOCK_SIZE,), dtype=tl.float32)
        for start_n in tl.range(0, N, BLOCK_SIZE):
            cols = start_n + base_cols
            mask = cols < N

            tl.store(dlogits_ptr+cols, dlogits.to(dlogits_ptr.dtype.element_ty), mask=mask)


class GrpoLoss(torch.autograd.Function):

    @input_guard
    @staticmethod
    def forward(ctx, logits, ref_logp, input_ids, advantages, beta, completion_mask, save_kl, inplace=True):
        ctx.input_shape = logits.shape
        B, L_ADD_1, N = ctx.input_shape
        L = L_ADD_1 - 1
        M = B * L
        input_ids_start_index = input_ids.size(1) - L

        if not save_kl:
            loss = torch.empty(B, L, device=logits.device, dtype=torch.float32)
        else:
            loss = torch.empty(B*2, L, device=logits.device, dtype=torch.float32)

        lse = torch.empty(B, L, device=logits.device, dtype=torch.float32)

        if completion_mask is None:
            completion_mask = torch.ones(B, L, device=logits.device, dtype=torch.int32)
        else:
            loss[:B].masked_fill_(completion_mask.logical_not(), 0.0)

        grpo_fwd_kernel[(M,)](
            logits_ptr=logits,
            ref_logp_ptr=ref_logp,
            input_ids_ptr=input_ids,
            advantages_ptr=advantages,
            completion_mask_ptr=completion_mask,
            loss_ptr=loss,
            lse_ptr=lse,
            beta=beta,
            save_kl=save_kl,
            B=B, M=M, N=N, L=L,
            start_idx=input_ids_start_index,
        )
        ctx.beta = beta
        ctx.save_for_backward(lse, logits, input_ids, advantages, completion_mask)
        ctx.ref_logp = ref_logp
        ctx.inplace = inplace
        return loss

    @input_guard
    @staticmethod
    def backward(ctx, dloss):
        # The grad of logits comes from two parts, the reward part and the kl part
        lse, logits, input_ids, advantages, completion_mask = ctx.saved_tensors
        inplace = ctx.inplace
        B, L_ADD_1, N = ctx.input_shape
        L = L_ADD_1 - 1
        M = B * L

        input_ids_start_index = input_ids.size(1) - L

        # B, L_ADD_1, N
        dlogits = logits if inplace else torch.empty_like(logits)
        BN = min(65536, triton.next_power_of_2(N))

        grpo_bwd_kernel[(M,)](
            dloss_ptr=dloss,
            dlogits_ptr=dlogits,
            logits_ptr=logits,
            ref_logp_ptr=ctx.ref_logp,
            input_ids_ptr=input_ids,
            advantages_ptr=advantages,
            completion_mask_ptr=completion_mask,
            lse_ptr=lse,
            beta=ctx.beta,
            B=B, N=N, L=L,
            BLOCK_SIZE=BN,
            start_idx=input_ids_start_index,
        )
        # The last token in the completion is not used in the loss computation
        # and therefore its gradient should be set to 0
        dlogits[:, -1, :].fill_(0.0)
        return dlogits.view(*ctx.input_shape), None, None, None, None, None, None, None


def fused_grpo_loss(logits, ref_logp, input_ids, advantages,
                    beta=0.1, completion_mask=None, save_kl=False, inplace=False) -> torch.Tensor:
    '''
    compute grpo loss, save memory(no addition usage) and fast speed(6X for A800)

    Args:
        logtits: Tensor, [B, L+1, vocab_size], the origin output of model, it's not logits[:, :-1]
        ref_logp: Tensor, [B, L], the origin output of model, it's not ref_logits[:, :-1]
        input_ids: Tensor, [B, K+L], it's prompt_completion_id, it contains the prompt ids and output ids
        advantages: Tensor, [B], the advantages of each prompt
        beta: float, the weight of kl loss
        completion_mask: Tensor, loss mask
        save_kl: bool, if true will save kl

    Retutn:
        loss: Tensor, [B, L], the loss of grpo, it contains the advantage part and kl part

    NOTE: logits(ref_logits) is computed by these steps
        logits_to_keep = completion_ids.size(1)

        def get_per_token_logits(model, input_ids, attention_mask, logits_to_keep):
            # We add 1 to `logits_to_keep` because the last logits of the sequence is later excluded
            logits = model(
                input_ids=input_ids, attention_mask=attention_mask, logits_to_keep=logits_to_keep + 1
            ).logits
            return logits

        logits = get_per_token_logits(model, prompt_completion_ids, attention_mask, logits_to_keep)
    '''
    out = GrpoLoss.apply(logits, ref_logp, input_ids, advantages, beta, completion_mask, save_kl, inplace)
    if not save_kl:
        return out
    else:
        return out.chunk(2, axis=0)


def grpo_loss_torch(logits, ref_logp, input_ids, advantages, beta=0.1, completion_mask=None, save_kl=False):
    def get_log_probs(logits, input_ids):
        per_token_logps = []
        for logits_row, input_ids_row in zip(logits, input_ids[:, -logits.size(1):], strict=False):
            log_probs = logits_row.log_softmax(dim=-1)
            token_log_prob = torch.gather(log_probs, dim=1, index=input_ids_row.unsqueeze(1)).squeeze(1)
            per_token_logps.append(token_log_prob)
        return torch.stack(per_token_logps)

    logits = logits[:, :-1]
    per_token_logps = get_log_probs(logits, input_ids)
    ref_per_token_logps = ref_logp
    per_token_kl = torch.exp(ref_per_token_logps - per_token_logps) - (ref_per_token_logps - per_token_logps) - 1

    per_token_loss = torch.exp(per_token_logps - per_token_logps.detach()) * advantages.unsqueeze(1)
    per_token_loss = -(per_token_loss - beta * per_token_kl)
    if completion_mask is not None:
        per_token_loss *= completion_mask
        if save_kl:
            per_token_kl *= completion_mask
    return per_token_loss if not save_kl else (per_token_loss, per_token_kl)


@torch.compile(fullgraph=True)
def grpo_loss_with_old_logps(
    logps: torch.Tensor,
    ref_logps: torch.Tensor,
    old_logps: torch.Tensor,
    pad_mask: torch.Tensor,
    logits_to_keep: int,
    rewards: torch.Tensor,
    beta: float = 0.2,
    epsilon: float = 0.2,
):
    """
    Compute the GRPO (Group Relative Policy Optimization) loss.

    Args:
        logps (torch.Tensor): [Batch, Token_length] Log probabilities of the current policy.
        ref_logps (torch.Tensor):[Batch, Token_length]  Log probabilities of the reference policy.
        old_logps (torch.Tensor): [Batch, Token_length] Log probabilities of the old policy.
        completion_ids (torch.Tensor): [Batch, Token_length] Completion token IDs (bool).
        pad_token_id: Pad token ID.
        logits_to_keep (int): Number of logits to keep for masking.
        rewards (torch.Tensor): [Batch] Rewards for each generation.
        beta (float) = 0.2: A hyperparameter for weighting the KL divergence term.
        epsilon (float) = 0.2: An float hyperparameter for clipping the importance weights.

    Returns:
        torch.Tensor: The computed GRPO loss.
    """
    B = logps.shape[0]
    assert B > 1, "Batch * Num generations should be greater than 1"

    rewards_shaped = rewards.view(-1, B)  # B,num_generations
    advantages = (rewards_shaped - rewards_shaped.mean(dim=1, keepdim=True)) / \
        (rewards_shaped.std(dim=1, keepdim=True) + 1e-8)
    advantages = advantages.view(-1)  # B*num_generations
    # Calculate the per - token KL divergence
    per_token_kl = torch.exp(ref_logps - logps) - (ref_logps - logps) - 1

    # Calculate the ratio of probabilities (importance weights)
    # Importance weights are calculated as exp(log_pi_theta - log_pi_theta_old)
    importance_weights = torch.exp(logps - old_logps)

    # Clip the importance weights to the range [1 - epsilon, 1 + epsilon]
    importance_weights_clipped = torch.clamp(importance_weights, 1 - epsilon, 1 + epsilon)

    # Create a completion mask. It checks which positions are valid based on logits_to_keep
    completion_mask = torch.arange(logits_to_keep, device=logps.device)[None, :] >= 0

    # Combine the completion mask and padding mask
    completion_mask = completion_mask & pad_mask  # Ensure matching shape

    # Add an extra dimension to advantages to match the shape for element - wise multiplication
    advantages = advantages.unsqueeze(1)

    # Calculate the per - token loss. It takes the minimum of the unclipped and clipped importance weights
    # and subtracts the KL divergence term weighted by beta, then multiplies by the completion mask
    token_loss = -(torch.min(advantages * importance_weights, advantages *
                   importance_weights_clipped) - beta * per_token_kl) * completion_mask

    # Calculate the final loss by summing the token losses and normalizing by the number of valid tokens
    loss = -token_loss.sum() / completion_mask.sum()

    return loss
